# -*- coding: utf-8 -*-

from typing import List
from pathlib import Path
from itertools import chain

from ..modules import my_classes as mc
from . import pre_survey_core as psc

__all__ = ["pre_survey", "default_pre_survey_param_dict", "pre_survey_separate_paths_input"]

default_pre_survey_param_dict = {
    'gap_open_penalty':     3, 
    'gap_extend_penalty':   1, 
    'match_score':          1, 
    'mismatch_score':       -2, 
    'distance_threshold':   5, 
    'number_of_groups':     1, 
}

def pre_survey(plasmid_map_dir_path: str, save_dir_base: str, ref_seq_aliases: List[str]=None, **param_dict: dict):
    plasmid_map_paths = [path for path in Path(plasmid_map_dir_path).glob("*.*") if path.suffix in mc.MyRefSeq.allowed_plasmid_map_extensions]
    return pre_survey_separate_paths_input(plasmid_map_paths, save_dir_base, ref_seq_aliases, **param_dict)

def pre_survey_separate_paths_input(plasmid_map_paths:List[Path], save_dir_base: str, ref_seq_aliases: List[str]=None, **param_dict:dict):
    for k in param_dict.keys():
        if k not in default_pre_survey_param_dict.keys():
            raise Exception(f"unknown key in `param_dict`: {k}\nallowed keys are: {', '.join(default_pre_survey_param_dict.keys())}")
    param_dict = {key: param_dict.get(key, val) for key, val in default_pre_survey_param_dict.items()}
    mc.assert_param_dict(param_dict)
    if len(plasmid_map_paths) == 0:
        raise Exception(f"Error: No plasmid map file was detected!")
    if param_dict["number_of_groups"] > len(plasmid_map_paths):
        raise Exception(f"Error: `number_of_groups` cannot be greater than the number of plasmid maps ({len(plasmid_map_paths)})!")
    # 1. Prepare objects
    ref_seq_list = [mc.MyRefSeq(plasmid_map_path) for plasmid_map_path in plasmid_map_paths]
    save_dir = mc.new_dir_path_wo_overlap(Path(save_dir_base) / Path(psc.RecommendedGrouping.file_name).stem, spacing="_")
    save_dir.mkdir()
    # 2. execute
    recommended_grouping = psc.execute_grouping(ref_seq_list, param_dict, save_dir, ref_seq_aliases)
    # 3. save results
    psc.export_results(recommended_grouping, save_dir)
    # 4. display recommended_grouping in the std output
    print(f"\n{recommended_grouping.recommended_grouping_txt}")
    return save_dir




# * coding: utf8 *

#@title # 1. Upload and select files

app_name = "SAVEMONEY"
version = "0.2.0"
description = "written by MU"

# total number of reads に対する insertion の割合が一定以上（2割とか3割とか？）になった場合を danger_zone と設定
# total number of reads に対する empty     の割合が一定以上（2割とか3割とか？）になった場合を empty_zone と設定
# danger_zone の周囲もdangerであるかを調べる

import re
import parasail
import numpy as np
import scipy.stats as stats

from . import my_classes as mc
from .cython_functions import alignment_functions as af

class MyAlignerBase(mc.MyCigarBase):
    alignment_algorithm_version = "aa_0.2.1"
    def __init__(self, param_dict):
        # params
        self.param_dict = param_dict
        self.gap_open_penalty = param_dict["gap_open_penalty"]
        self.gap_extend_penalty = param_dict["gap_extend_penalty"]
        self.match_score = param_dict["match_score"]
        self.mismatch_score = param_dict["mismatch_score"]
    @property
    def my_custom_matrix(self):
        return parasail.matrix_create("ACGT", self.match_score, self.mismatch_score)
    def my_special_sw(self, ref_seq, query_seq):
        result = parasail.sw_trace(query_seq, ref_seq, self.gap_open_penalty, self.gap_extend_penalty, self.my_custom_matrix)
        my_result = MyResult(result)
        my_result.my_cigar = (
            "H" * my_result.beg_ref + 
            "S" * my_result.beg_query + 
            my_result.my_cigar + 
            "S" * (len(query_seq) - my_result.end_query - 1) + 
            "H" * (len(ref_seq) - my_result.end_ref - 1)
        )
        my_result.beg_query = 0
        my_result.end_query = len(query_seq) - 1
        return my_result
    def nw_trace(self, ref_seq, query_seq):
        result = parasail.nw_trace(query_seq, ref_seq, self.gap_open_penalty, self.gap_extend_penalty, self.my_custom_matrix)
        return MyResult(result)
    def my_special_sg_qe_de(self, ref_seq, query_seq):
        """
        ATCGATCGATCGATCG
        ATCGATCG--------
        or
        ATCGATCG--------
        ATCGATCGATCGATCG
        """
        result = parasail.sg_qe_de_trace(query_seq, ref_seq, self.gap_open_penalty, self.gap_extend_penalty, self.my_custom_matrix)
        my_result = MyResult(result)
        assert my_result.beg_ref == 0
        assert my_result.beg_query == 0
        my_result.set_soft_clipping(self.gap_open_penalty, self.gap_extend_penalty, self.match_score, self.mismatch_score, side="beg")
        return my_result
    def my_special_sg_qb_db(self, ref_seq, query_seq):
        """
        ATCGATCGATCGATCG
        --------ATCGATCG
        or
        --------ATCGATCG
        ATCGATCGATCGATCG
        """
        result = parasail.sg_qb_db_trace(query_seq, ref_seq, self.gap_open_penalty, self.gap_extend_penalty, self.my_custom_matrix)
        my_result = MyResult(result)
        assert my_result.end_ref == len(ref_seq) - 1
        assert my_result.end_query == len(query_seq) - 1
        my_result.set_soft_clipping(self.gap_open_penalty, self.gap_extend_penalty, self.match_score, self.mismatch_score, side="end")
        return my_result
    # query の間はギャップペナルティを与えないアラインメント
    def my_special_DP(self, query_seq_1: str, query_seq_2: str, ref_seq: str):
        ################
        # アラインメント #
        ################
        if len(query_seq_1) > 0:
            my_result_1 = self.my_special_sg_qe_de(ref_seq, query_seq_1)
        else:
            my_result_1 = MyResult()
            my_result_1.my_cigar = "H" * len(ref_seq)
            my_result_1.beg_ref = -1
            my_result_1.end_ref = -1
        if len(query_seq_2) > 0:
            my_result_2 = self.my_special_sg_qb_db(ref_seq, query_seq_2)
        else:
            my_result_2 = MyResult()
            my_result_2.my_cigar = "H" * len(ref_seq)
            my_result_2.beg_ref = len(ref_seq)
            my_result_2.end_ref = len(ref_seq)
        ########
        # 結合 #
        ########
        if my_result_2.beg_ref - my_result_1.end_ref - 1 > 0:
            """
            ref         ?????????????????????     ??????         ?????????????????????
            my_cigar_1  =====ALIGNED-SEQ=====SSSSSHHHHHH         HHHHHHHHHHHHHHHHHHHHH
            my_cigar_2  HHHHHHHHHHHHHHHHHHHHH     HHHHHHSSSSSSSSS=====ALIGNED-SEQ=====
            combined    =====ALIGNED-SEQ=====SSSSSHHHHHHSSSSSSSSS=====ALIGNED-SEQ=====
            """
            len_H = my_result_2.beg_ref - my_result_1.end_ref - 1
            my_result = MyResult()
            my_result.my_cigar = (
                my_result_1.my_cigar[:len(my_result_1.my_cigar) - my_result_1.my_cigar.count("H")] + 
                "H" * len_H + 
                my_result_2.my_cigar[my_result_2.my_cigar.count("H"):]
            )
        else:
            """ ベストな区切りを探すよ
            ref         ??????????????????????????????????????????????????
            my_cigar_1  =====ALIGNED-SEQ===========SSSSSHHHHHHHHHHHHHHHHHH
            my_cigar_2  HHHHHHHHHHHHSSSSSSSSS=============ALIGNED-SEQ=====
            """
            from . import msa
            my_msa = msa.MyMSA.generate_msa(mc.MySeq(ref_seq), [query_seq_1, query_seq_2], q_scores_list=[[-1] * len(query_seq_1), [-1] * len(query_seq_2)], my_cigar_list=[my_result_1.my_cigar, my_result_2.my_cigar], param_dict=self.param_dict)
            # find best position to cut
            my_cigar_1_aligned, my_cigar_2_aligned = my_msa.my_cigar_list_aligned   # 逆端に O が入ることはあっても H が入ることはない
            cumsum_scores_1_from_left = self.cumsum_my_cigar_scores(my_cigar_1_aligned)
            cumsum_scores_2_from_right = self.cumsum_my_cigar_scores(my_cigar_2_aligned[::-1])[::-1]
            idx_to_cut = (np.array(cumsum_scores_1_from_left) + np.array(cumsum_scores_2_from_right)).argmax()
            len_S_1 = (
                len(cumsum_scores_1_from_left) - 1 - idx_to_cut - 
                my_cigar_1_aligned[idx_to_cut:].count("D") - 
                my_cigar_1_aligned[idx_to_cut:].count("N") - 
                my_cigar_1_aligned[idx_to_cut:].count("H") - 
                my_cigar_1_aligned[idx_to_cut:].count("O")
            )
            len_S_2 = (
                idx_to_cut - 
                my_cigar_2_aligned[:idx_to_cut].count("D") - 
                my_cigar_2_aligned[:idx_to_cut].count("N") - 
                my_cigar_2_aligned[:idx_to_cut].count("H") - 
                my_cigar_2_aligned[:idx_to_cut].count("O")
            )
            my_cigar = my_cigar_1_aligned[:idx_to_cut] + "S" * len_S_1 + "S" * len_S_2 + my_cigar_2_aligned[idx_to_cut:]
            my_cigar = my_cigar.replace("N", "").replace("O", "")
            """ # assert my_cigar.count("H") == 0 は不要: 例えば下記の場合に H が入る
            ref_seq     ACTTGTTTAGAAGA    AATAAT GGAA
            query_seq_1 ACTTGTTTAGAAGAGATG
            my_cigar_1  ==============SSSSHHHHHHOHHHH
            query_seq_2 ACTTGTTT    GA    GATAATGGGAG
            my_cigar_2  ========DDDD==NNNNX=====I===X
            combined    ==============SSSSH=====I===X
            """
            my_result = MyResult()
            my_result.my_cigar = my_cigar
        return my_result
    def cumsum_my_cigar_scores(self, my_cigar):
        cumsum_scores = [0]
        for LLL, L in self.generate_cigar_iter(my_cigar):
            N = len(LLL)
            if L == "=":
                cumsum_scores.extend(
                    np.arange(1, N + 1) * self.match_score + cumsum_scores[-1]
                )
            elif L == "X":
                cumsum_scores.extend(
                    np.arange(1, N + 1) * self.mismatch_score + cumsum_scores[-1]
                )
            elif L in "DI":
                s = cumsum_scores[-1] - self.gap_open_penalty
                cumsum_scores.extend(
                    [s] + list(s - np.arange(1, N) * self.gap_extend_penalty)
                )
            elif L in "SHON":
                cumsum_scores.extend([cumsum_scores[-1] for i in range(N)])
            else:
                raise Exception(f"error: {L}")
        return cumsum_scores

class MyOptimizedAligner(MyAlignerBase, mc.AlignmentBase):
    default_repeat_max = 5
    percentile_factor = 0.1 # 低いほど conserved region の quality は上がるが、計算時間がかかる
    omit_too_long = 1.75    # omit_too_long: 1.5-2.0 くらいの値: self.calc_circular_conserved_region の elif 文参照
    def __init__(self, ref_seq, param_dict) -> None:
        super().__init__(param_dict)
        self.ref_seq = ref_seq
        self.ref_seq_v = np.array([ord(char) for char in ref_seq])
        self.N_ref = len(self.ref_seq_v)
        self.ref_seq_v_repeated_list = [
            np.hstack([self.ref_seq_v for i in range(repeat)]).astype(np.int64)
            for repeat in range(2, self.default_repeat_max + 1)
        ]
        self.k_list = np.arange(0, self.N_ref)
        x_percentile = 1 / self.N_ref * self.percentile_factor # 偶然 threshold を超える position が 0.1個/ref_seq 以下になるようにする
        self.sigma = stats.norm.ppf(1 - x_percentile, loc=0, scale=1)  # 累積分布関数の逆関数
        self.is_all_ATCG = all([b.upper() in "ATCG" for b in ref_seq])
    def calc_circular_conserved_region(self, query_seq, omit_too_long=False):
        conserved_regions = self.calc_circular_conserved_region_core(query_seq)
        if len(conserved_regions) == 0:
            return None
        elif omit_too_long and (self.N_ref * self.omit_too_long <= len(query_seq)):
            return None
        else:
            longest_path_trace, longest_path_score = SearchLongestPath.exec(conserved_regions, N_ref=self.N_ref, N_query=len(query_seq))
            # 順序づけ
            conserved_regions.tidy_conserved_regions(longest_path_trace)
            return conserved_regions
    def execute_circular_alignment_using_conserved_regions(self, query_seq, conserved_regions):
        N_query = len(query_seq)
        # set offset
        query_end_idx_after_offset = conserved_regions.set_tidy_offset(len(self.ref_seq), N_query)   # ラストの non_conserved_region もしくは ラストの conserved_region の末端に query_end_idx_after_offset が含まれることを保証する
        self.ref_seq.set_offset(conserved_regions.ref_seq_offset)
        query_seq.set_offset(conserved_regions.query_seq_offset)
        # execute alignment for non_conserved_regiong
        result_master = MyResult()
        for length_of_previous_conserved_region, non_conserved_ref_start, non_conserved_ref_end, non_conserved_query_start, non_conserved_query_end in conserved_regions.iter_regions():
            # print(length_of_previous_conserved_region, non_conserved_ref_start, non_conserved_ref_end, non_conserved_query_start, non_conserved_query_end)
            # 最後のループはは常に -1 になっているはず (conserved_regions.iter_regions() の中で assert している)
            if non_conserved_ref_end != -1:
                # non_conserved_ref の長さが 0 の場合
                if non_conserved_ref_start > non_conserved_ref_end:
                    assert non_conserved_ref_start - 1 == non_conserved_ref_end
                    result_master.append_my_cigar("=" * length_of_previous_conserved_region)
                    result_master.append_my_cigar("I" * (non_conserved_query_end - non_conserved_query_start + 1))
                    continue
                # non_conserved_query の長さが 0 の場合（non_conserved_ref, non_conserved_query の長さが同時に 0 になることがあるが、多分大丈夫なはず）
                elif non_conserved_query_start > non_conserved_query_end:
                    assert non_conserved_query_start - 1 == non_conserved_query_end
                    result_master.append_my_cigar("=" * length_of_previous_conserved_region)
                    result_master.append_my_cigar("D" * (non_conserved_ref_end - non_conserved_ref_start + 1))
                    continue
                else:
                    ref_seq_extracted = self.ref_seq[non_conserved_ref_start:non_conserved_ref_end+1]
                    query_seq_extracted = query_seq[non_conserved_query_start:non_conserved_query_end+1]
                    # execute alignment # non_conserved_region の両端は完全一致している conserved_region のはずなので、ローカルアラインメントではない！
                    result = parasail.nw_trace(query_seq_extracted, ref_seq_extracted, self.gap_open_penalty, self.gap_extend_penalty, self.my_custom_matrix)
                    my_result = MyResult(result)
            # ラストループ
            else:
                assert non_conserved_query_end == -1
                # non_conserved_ref の長さが 0 の場合
                if non_conserved_ref_start == self.N_ref:
                    result_master.append_my_cigar("=" * length_of_previous_conserved_region)
                    result_master.append_my_cigar("S" * (N_query - non_conserved_query_start))  # ソフトクリップ
                    continue
                # non_conserved_query の長さが 0 の場合（non_conserved_ref, non_conserved_query の長さが同時に 0 になることがあるが、多分大丈夫なはず）
                elif non_conserved_query_start == N_query:
                    result_master.append_my_cigar("=" * length_of_previous_conserved_region)
                    result_master.append_my_cigar("H" * (self.N_ref - non_conserved_ref_start)) # ハードクリップ
                    continue
                else:
                    ref_seq_extracted = self.ref_seq[non_conserved_ref_start:]
                    query_seq_extracted = query_seq[non_conserved_query_start:]
                    # execute alignment
                    query_seq_extracted_1 = query_seq[non_conserved_query_start:query_end_idx_after_offset + 1]
                    query_seq_extracted_2 = query_seq[query_end_idx_after_offset + 1:]
                    my_result = self.my_special_DP(
                        query_seq_extracted_1, 
                        query_seq_extracted_2, 
                        ref_seq_extracted, 
                    )

            # 結果を追加
            result_master.append_my_cigar("=" * length_of_previous_conserved_region)
            result_master.append_my_cigar(my_result.my_cigar)

        # self.print_alignment(self.ref_seq, query_seq, result_master.my_cigar)

        # ref_seq_offset が 0 になるような感じで、queyr_seq_offset を調整しよう (conserved_regions.query_seq_offset の場合は conserved_regions.ref_seq_offset != 0 であることがほとんど)
        new_query_seq_offset = result_master.set_ref_seq_offset_as_0(conserved_regions.ref_seq_offset, conserved_regions.query_seq_offset)
        result_master.set_cigar_score(self.gap_open_penalty, self.gap_extend_penalty, self.match_score, self.mismatch_score)

        # assertion
        self.ref_seq.set_offset(0)
        query_seq.set_offset(new_query_seq_offset)
        self.assert_alignment(self.ref_seq, query_seq, result_master.my_cigar)

        # prepare for nexe analysis
        query_seq.set_offset(0)

        # プラスミドダイマーの場合
        # if result_master.score < 0:
        #     # view alignment
        #     print()
        #     self.ref_seq.set_offset(0)
        #     query_seq.set_offset(new_query_seq_offset)
        #     self.print_alignment(self.ref_seq, query_seq, result_master.my_cigar)
        #     quit()
        return result_master
    def calc_circular_conserved_region_core(self, query_seq):
        # 文字列を数字に変換
        query_seq_v = np.array([ord(char) for char in query_seq], dtype=np.int64)
        N_query = len(query_seq_v)
        repeat = 1 + np.ceil((N_query - 1) / self.N_ref).astype(int)
        # アラインメントのたびに ref_seq_v_repeated を作らないようにする：本来は else 中の hstack のみで ok だけど、それだと時間がかかるかな？
        if repeat <= self.default_repeat_max:
            ref_seq_v_repeated = self.ref_seq_v_repeated_list[repeat - 2]
        else:
            ref_seq_v_repeated = np.hstack([self.ref_seq_v for i in range(repeat)]).astype(np.int64)

        #######################
        # k座位分だけ query をずらして並べたときの類似性（repeated ref_seq を固定して、query_seq のオフセットを右にずらしていくイメージ）
        #######################
        """
        # python version
        """
        # result_array = [(ref_seq_v_repeated[k:k + N_query] == query_seq_v).sum() for k in self.k_list]
        """
        # cython version 1
        """
        # result_array = np.empty_like(self.k_list, dtype=np.int64)
        # af.k_mer_offset_analysis(ref_seq_v_repeated, query_seq_v, result_array)
        """
        # cython version 2 (0.012740850448608398s)
        """
        result_array = af.k_mer_offset_analysis_2(ref_seq_v_repeated, query_seq_v, self.N_ref, len(query_seq_v))

        # 閾値を超える類似性が得られた場合の k座位を取得 (0.001313924789428711s)
        selected_k_list = np.where(result_array > np.median(result_array) + self.sigma * np.std(result_array))[0]

        # import matplotlib.pyplot as plt
        # plt.plot(self.k_list, result_array)
        # plt.axhline(np.median(result_array), color="r")
        # plt.axhline(np.median(result_array) + self.sigma * np.std(result_array), color="r")
        # plt.axhline(np.median(result_array) - self.sigma * np.std(result_array), color="r")
        # plt.show()

        # ウィンドウ解析
        conserved_region_start_idx_list_query = []
        conserved_region_end_idx_list_query = []
        conserved_region_start_idx_list_ref = []
        conserved_region_end_idx_list_ref = []

        for i, k in enumerate(selected_k_list):
            start_idx_sublist, end_idx_sublist = WindowAnalysis.exec(ref_seq_v_repeated[k:k + N_query], query_seq_v)
            conserved_region_start_idx_list_ref.extend(start_idx_sublist + k)
            conserved_region_end_idx_list_ref.extend(end_idx_sublist + k)
            conserved_region_start_idx_list_query.extend(start_idx_sublist)
            conserved_region_end_idx_list_query.extend(end_idx_sublist)
        conserved_regions_ref = np.vstack((conserved_region_start_idx_list_ref, conserved_region_end_idx_list_ref)).T
        conserved_regions_ref %= self.N_ref # リピートが 3 以上の場合は、複数回引く必要がある：でないと index エラーになる
        conserved_regions_query = np.vstack((conserved_region_start_idx_list_query, conserved_region_end_idx_list_query)).T
        return ConservedRegions(np.hstack((conserved_regions_ref, conserved_regions_query)))

class MyResult(mc.MyCigarBase):
    def __init__(self, parasail_result=None) -> None:
        if parasail_result is not None:
            self.my_cigar = self.cigar_to_my_cigar(parasail_result.cigar.decode.decode("ascii"))
            self.score = parasail_result.score
            self.beg_ref = parasail_result.cigar.beg_ref
            self.beg_query = parasail_result.cigar.beg_query
            self.end_ref = parasail_result.end_ref
            self.end_query = parasail_result.end_query
        else:
            self.my_cigar = ""
            self.score = 0
            self.beg_ref = 0
            self.beg_query = 0
            self.end_ref = 0
            self.end_query = 0
        # ref_seq_offset = 0 となるようなアラインメントが行われた時にのみ値が登録される (see self.set_ref_seq_offset_as_0)
        self.new_query_seq_offset = None
    def __str__(self) -> str:
        return f"cigar\t{self.my_cigar}\nscore\t{self.score}\nref\t{self.beg_ref}\t{self.end_ref}\nquery\t{self.beg_query}\t{self.end_query}\nnew_query_seq_offset\t{self.new_query_seq_offset}"
    def __len__(self):
        return len(self.my_cigar)
    def append_my_cigar(self, my_cigar):
        self.my_cigar += my_cigar
    @property
    def cigar(self):
        return "".join(f"{len(LLL)}{L}" for LLL, L in self.generate_cigar_iter(self.my_cigar))
    def set_cigar_score(self, gap_open_penalty, gap_extend_penalty, match_score, mismatch_score):
        self.score = 0
        for LLL, L in self.generate_cigar_iter(self.my_cigar):
            if L == "=":
                self.score += match_score * len(LLL)
            elif L == "X":
                self.score += mismatch_score * len(LLL)
            elif L in "DI":
                self.score -= gap_open_penalty + gap_extend_penalty * (len(LLL) - 1)
            elif L in "SHE":
                pass
            else:
                raise Exception(f"unknown letter code: {L}")
    def set_ref_seq_offset_as_0(self, ref_seq_offset, query_seq_offset):
        aligned_ref_seq_offset = self.get_aligned_ref_seq_offset(ref_seq_offset)
        aligned_query_seq_offset = self.get_aligned_query_seq_offset(query_seq_offset)
        diff = aligned_query_seq_offset - aligned_ref_seq_offset
        if aligned_query_seq_offset > aligned_ref_seq_offset:
            if aligned_ref_seq_offset != 0:
                new_query_seq_offset = diff - self.my_cigar[-aligned_query_seq_offset:-aligned_ref_seq_offset].count("D") - self.my_cigar[-aligned_query_seq_offset:-aligned_ref_seq_offset].count("H")
            else:
                new_query_seq_offset = diff - self.my_cigar[-aligned_query_seq_offset:].count("D") - self.my_cigar[-aligned_query_seq_offset:].count("H")
        else:
            # aligned_query_seq_offset != 0 の場合
            if aligned_query_seq_offset != 0:
                new_query_seq_offset = diff + self.my_cigar[-aligned_ref_seq_offset:-aligned_query_seq_offset].count("D") + self.my_cigar[-aligned_ref_seq_offset:-aligned_query_seq_offset].count("H")
            # aligned_ref_seq_offset != aligned_query_seq_offset == 0 の場合
            elif aligned_ref_seq_offset != 0:
                new_query_seq_offset = diff + self.my_cigar[-aligned_ref_seq_offset:].count("D") + self.my_cigar[-aligned_ref_seq_offset:].count("H")
            # aligned_ref_seq_offset == aligned_query_seq_offset == 0 の場合
            else:
                new_query_seq_offset = 0
        # 自身のアップデート
        self.my_cigar = self.my_cigar[-aligned_ref_seq_offset:] + self.my_cigar[:-aligned_ref_seq_offset]
        self.new_query_seq_offset = new_query_seq_offset
        return new_query_seq_offset
    def get_aligned_ref_seq_offset(self, ref_seq_offset):
        if ref_seq_offset == 0:
            return 0
        assert ref_seq_offset > 0
        aligned_ref_seq_offset = ref_seq_offset
        previous_N_ins = 0
        N_ins = self.my_cigar[-aligned_ref_seq_offset:].count("I") + self.my_cigar[-aligned_ref_seq_offset:].count("S")
        while ref_seq_offset != aligned_ref_seq_offset - N_ins:
            aligned_ref_seq_offset += N_ins - previous_N_ins
            previous_N_ins = N_ins
            N_ins = self.my_cigar[-aligned_ref_seq_offset:].count("I") + self.my_cigar[-aligned_ref_seq_offset:].count("S")
        return aligned_ref_seq_offset
    def get_aligned_query_seq_offset(self, query_seq_offset):
        if query_seq_offset == 0:
            return 0
        assert query_seq_offset > 0
        aligned_query_seq_offset = query_seq_offset
        previous_N_del = 0
        N_del = self.my_cigar[-aligned_query_seq_offset:].count("D") + self.my_cigar[-aligned_query_seq_offset:].count("H")
        while query_seq_offset != aligned_query_seq_offset - N_del:
            aligned_query_seq_offset += N_del - previous_N_del
            previous_N_del = N_del
            N_del = self.my_cigar[-aligned_query_seq_offset:].count("D") + self.my_cigar[-aligned_query_seq_offset:].count("H")
        return aligned_query_seq_offset
    def to_dict(self):
        d = {
            "cigar":self.cigar,     # 出力容量節約の為、my_cigarr でなく、cigar で出力する
            "score":self.score, 
            "new_query_seq_offset":self.new_query_seq_offset, 
        }
        return d
    def apply_dict_params(self, d):
        setattr(self, "score", int(d["score"]))
        setattr(self, "my_cigar", self.cigar_to_my_cigar(d["cigar"]))
        new_query_seq_offset = d["new_query_seq_offset"]
        if new_query_seq_offset == "None":
            setattr(self, "new_query_seq_offset", None)
        else:
            setattr(self, "new_query_seq_offset", int(new_query_seq_offset))
    # for pre-survey
    def calc_levenshtein_distance(self):
        return len(self.my_cigar) - self.my_cigar.count("=")
    # # for semi-global alignment
    # def organize_end(self):
    #     assert self.beg_ref == self.beg_query == 0
    #     m = re.search(r"^[=XDI]+?(I*)(D*)$", self.my_cigar)
    #     len_endI = len(m.group(1))
    #     len_endD = len(m.group(2))
    #     if len_endD == 0:
    #         len_endID = len_endI
    #         end_cigar = "S"
    #     else:
    #         len_endID = len_endD
    #         end_cigar = "H"
    #     alignment_len = len(self.my_cigar) - len_endID
    #     self.end_ref = alignment_len - self.my_cigar[:alignment_len].count("I") - 1
    #     self.end_query = alignment_len - self.my_cigar[:alignment_len].count("D") - 1
    #     assert self.end_ref >= 0
    #     assert self.end_query >= 0
    #     self.my_cigar = self.my_cigar[:alignment_len] + end_cigar * len_endID
    # def organize_beg(self):
    #     m = re.search(r"^(D*)(I*)[=XDI]+?$", self.my_cigar)
    #     len_begI = len(m.group(1))
    #     len_begD = len(m.group(2))
    #     if len_begD == 0:
    #         len_begID = len_begI
    #         end_cigar = "S"
    #     else:
    #         len_begID = len_begD
    #         end_cigar = "H"
    #     self.beg_ref = len_begID - self.my_cigar[:len_begID].count("I")
    #     self.beg_query = len_begID - self.my_cigar[:len_begID].count("D")
    #     self.my_cigar = "H" * self.beg_ref + self.my_cigar[self.beg_ref:]
    #     return self.beg_ref # = len_begD
    def set_soft_clipping(self, gap_open_penalty, gap_extend_penalty, match_score, mismatch_score, side):
        if side == "end":
            my_cigar = self.my_cigar[::-1]
        elif side == "beg":
            my_cigar = self.my_cigar
        else:
            raise Exception(f"error: {side}")
        my_cigar, len_aligned_ref = self.set_soft_clipping_core(my_cigar, gap_open_penalty, gap_extend_penalty, match_score, mismatch_score)
        if side == "end":
            self.my_cigar = my_cigar[::-1]
            self.beg_ref = my_cigar.count("H")
            self.beg_query = 0
        elif side == "beg":
            self.my_cigar = my_cigar
            self.end_ref = len_aligned_ref - 1
            self.end_query = len(my_cigar) - my_cigar.count("H") - my_cigar.count("D") - 1
    @staticmethod
    def set_soft_clipping_core(my_cigar, gap_open_penalty, gap_extend_penalty, match_score, mismatch_score):
        """ 下記のようなイメージにする
        ### PRE ###
        cigar   ==I=DX===XXXXXDDD
        ref     AA-AAAAAATTTTTTTT
        query   AAAA-TAAACCCCC

        ### POST ###
        cigar   ==I=DX===SSSSSHHHHHHHH
        ref     AA-AAAAAA     TTTTTTTT
        query   AAAA-TAAACCCCC
        """
        score_list = []
        cur_score = 0
        previous_L = ""
        for L in my_cigar:
            if L == "=":
                cur_score += match_score
            elif L == "X":
                cur_score += mismatch_score
            elif L in "ID":
                if previous_L == L:
                    cur_score -= gap_extend_penalty
                else:
                    cur_score -= gap_open_penalty
            else:
                raise Exception(f"error: {L}")
            score_list.append(cur_score)
            previous_L = L
        max_score_idx = np.argmax(score_list)
        len_aligned = max_score_idx + 1
        len_aligned_and_preS = len(score_list)
        len_aligned_and_preH = len(my_cigar)
        if score_list[max_score_idx] < 0:
            len_query = len_aligned_and_preS - my_cigar[:len_aligned_and_preS].count("D")
            len_ref = len_aligned_and_preH - my_cigar.count("I")
            new_my_cigar = "S" * len_query  + "H" * len_ref
            len_aligned_ref = 0
        else:
            len_query_S = len_aligned_and_preS - len_aligned - my_cigar[len_aligned:len_aligned_and_preS].count("D")
            len_ref_H = len_aligned_and_preH - len_aligned - my_cigar[len_aligned:].count("I")
            new_my_cigar = my_cigar[:len_aligned] + "S" * len_query_S + "H" * len_ref_H
            len_aligned_ref = len_aligned - my_cigar[:len_aligned].count("I")
        return new_my_cigar, len_aligned_ref
    ##############################
    # Used for polishing process #
    ##############################
    def get_mutated_regions(self):
        return [(m.group(0), m.start(), m.end() - 1, self.my_cigar[:m.start()].count("I"), self.my_cigar[:m.start()].count("D")) for m in re.finditer(r"([IDX]+)", self.my_cigar)]
    def get_corresponding_query_idx(self, ref_idx):
        cur_query_idx = -1
        cur_ref_idx = -1
        for c in self.my_cigar:
            if c in "=X":
                cur_ref_idx += 1
                cur_query_idx += 1
            elif c == "I":
                cur_query_idx += 1
            elif c == "D":
                cur_ref_idx += 1
            else:
                raise Exception(f"error: {c}")
            if cur_ref_idx == ref_idx:
                if c == "I":
                    raise Exception(f"error {c}")
                else:
                    return cur_query_idx
        print(cur_ref_idx, cur_query_idx)
        raise Exception("error")

class ConnectionMatrix():
    def __init__(self, connection_matrix, initial_node_idx=0, initial_node_score=0) -> None:
        self.connection_matrix = np.copy(connection_matrix)
        self.initial_node_idx = initial_node_idx
        self.initial_node_score = initial_node_score

        self.N_node = self.connection_matrix.shape[0]
        self.longest_path_trace_list = [None for i in range(self.N_node)]
        self.longest_path_trace_list[self.initial_node_idx] = [self.initial_node_idx]
        self.longest_path_score_list = [0 for i in range(self.N_node)]
        self.longest_path_score_list[self.initial_node_idx] += self.initial_node_score

        self.connection_matrix_remained_to_be_analyzed = None
    def remove_unreachable_branches(self):
        # 便宜上数値を1つだけ適当な値を代入（このクラスメソッドの最後で元に戻されるよ）
        self.connection_matrix[self.initial_node_idx, self.initial_node_idx] = 100  # 値は何でも良いが、必要

        N_idx_with_no_input_previous = []
        update = True
        while update:
            # 入力が無いノードへは行けない
            idx_with_no_input = np.where(self.connection_matrix.sum(axis=0) == 0)[0]
            N_idx_with_no_input = len(idx_with_no_input)
            if N_idx_with_no_input == N_idx_with_no_input_previous:
                update = False
            else:
                for idx in idx_with_no_input:
                    self.connection_matrix[idx, :] = 0
                # 後処理
                N_idx_with_no_input_previous = N_idx_with_no_input

        # 後処理
        self.connection_matrix[self.initial_node_idx, self.initial_node_idx] = 0
    def execute_dijkstras_algorithm(self):
        self.connection_matrix_remained_to_be_analyzed = self.connection_matrix > 0
        newly_determined_idx_list = [self.initial_node_idx]
        while self.connection_matrix_remained_to_be_analyzed.any(axis=None):
            determined_idx_list = []
            for idx in newly_determined_idx_list:
                for col in np.where(self.connection_matrix[idx, :] > 0)[0]:
                    score_so_far = self.longest_path_score_list[idx]
                    acquired_score = self.connection_matrix[idx, col]
                    total_score = score_so_far + acquired_score
                    # コネクション (i, j) が、既存よりより長いパスであるようなら、更新
                    if total_score > self.longest_path_score_list[col]:
                        self.longest_path_score_list[col] = total_score
                        self.longest_path_trace_list[col] = self.longest_path_trace_list[idx] + [col]
                    # 済をチェック -> all済なら、determined_idx_list に登録
                    self.connection_matrix_remained_to_be_analyzed[idx, col] = False
                    if not any(self.connection_matrix_remained_to_be_analyzed[:, col]):
                        determined_idx_list.append(col)
            newly_determined_idx_list = determined_idx_list
        # 一番長いパスを取得
        max_i = np.argmax(self.longest_path_score_list)
        return self.longest_path_trace_list[max_i], self.longest_path_score_list[max_i]
    def ignore_nodes(self, node_index_to_ignore):
        for node_idx in node_index_to_ignore:
            self.connection_matrix[node_idx, :] = 0
            self.connection_matrix[:, node_idx] = 0
    def __str__(self) -> str:
        return str(self.connection_matrix)

class ConservedRegions():
    """
    self.conserved_regions.shape = (N, 4)  # ref_start, ref_end, query_start, query_end
    """
    def __init__(self, conserved_regions) -> None:
        assert conserved_regions.shape[1] == 4
        self.conserved_regions = conserved_regions
        self.ref_seq_offset = 0
        self.query_seq_offset = 0
    def copy(self):
        return self.__class__(np.copy(self.conserved_regions))
    def tidy_conserved_regions(self, longest_path_trace):
        # re-order
        self.conserved_regions = self.conserved_regions[longest_path_trace, :]
    def set_tidy_offset(self, N_ref, N_query):
        assert len(self.conserved_regions) > 0
        assert all(self.conserved_regions[:, 2] >= 0)
        assert self.ref_seq_offset == self.query_seq_offset == 0
        assert self.conserved_regions[-1, 2] < self.conserved_regions[-1, 3]    # consered_regions の検出では ref を繰り返した配列を作成して query を動かしているので、queryの最初最後にまたがった conserved_region が作成されることはないはず
        # min(query_start) が、一番上の行にくるように並び替え -> これで、query の切れ目が最後の non_conserved_region に来ることが保証される
        self.conserved_regions = np.roll(self.conserved_regions, -self.conserved_regions[:, 2].argmin(), axis=0)
        # 最初の行を [0, x, 0, y] となるように offset を取る
        self.set_offset(self.conserved_regions[0, 0], self.conserved_regions[0, 2], N_ref, N_query)
        assert self.conserved_regions[0, 0] == self.conserved_regions[0, 2] == 0
        # query の切れ目がどこに属するかが返り値 (ref_end_idx_after_offset + 1 = ref_start_idx_after_offset であることに注意)
        return N_query - self.query_seq_offset - 1  # ref_end_idx_after_offset
        query_end_idx_after_offset = N_query - self.query_seq_offset - 1
        query_start_idx_after_offset = (N_query - self.query_seq_offset)%N_query
    def set_offset(self, ref_start, query_start, N_ref, N_query):
        self.ref_seq_offset = ref_start
        self.query_seq_offset = query_start
        self.conserved_regions[:, :2] -= self.ref_seq_offset
        self.conserved_regions[:, 2:] -= self.query_seq_offset
        # 負の index を補正
        below_zero_ref = self.conserved_regions < 0
        below_zero_query = np.copy(below_zero_ref)
        below_zero_ref[:, 2:] = False
        below_zero_query[:, :2] = False
        self.conserved_regions[below_zero_ref] += N_ref
        self.conserved_regions[below_zero_query] += N_query
    # @property
    # def non_conserved_regions(self):
    #     non_conserved_regions = np.empty_like(self.conserved_regions)
    #     for i, (ref_start, ref_end, query_start, query_end) in enumerate(self):
    #         non_conserved_regions[i, 0] = ref_end + 1
    #         non_conserved_regions[i, 2] = query_end + 1
    #         non_conserved_regions[i - 1, 1] = ref_start - 1
    #         non_conserved_regions[i - 1, 3] = query_start - 1
    #     return non_conserved_regions
    def iter_regions(self):
        """
        regions.shape = (N, 5)  # length_of_previous_conserved_region, non_conserved_ref_start, non_conserved_ref_end, non_conserved_query_start, non_conserved_query_end
        """
        regions = np.empty((self.conserved_regions.shape[0], 5), dtype=int)
        for i, (ref_start, ref_end, query_start, query_end) in enumerate(self):
            regions[i, 0] = ref_end - ref_start + 1
            regions[i, 1] = ref_end + 1
            regions[i, 3] = query_end + 1
            regions[i - 1, 2] = ref_start - 1
            regions[i - 1, 4] = query_start - 1
        assert regions[-1, 2] == regions[-1, 4] == -1
        return regions
    def __iter__(self):
        yield from self.conserved_regions
    def __str__(self) -> str:
        return self.conserved_regions.__str__()
    def __len__(self):
        return len(self.conserved_regions)

#####################
# メソッドのみのクラス #
#####################
class SearchLongestPath():
    @classmethod
    def exec(cls, conserved_regions:ConservedRegions, N_ref:int, N_query:int):   # conserved_regions.shape = (N, 4)  # ref_start, ref_end, query_start, query_end
        longest_path_trace_list = []
        longest_path_score_list = []
        # 各々の conserved_region を始点とした時の、最適スコアを計算する
        for ref_start, ref_end, query_start, query_end in conserved_regions.conserved_regions:
            # 各 conserved_region を始点とするように offset を調整して、簡易DPを実行
            conserved_regions_copied = conserved_regions.copy()
            conserved_regions_copied.set_offset(ref_start, query_start, N_ref, N_query)
            longest_path_trace, longest_path_score = cls.search_longest_path(conserved_regions_copied)
            longest_path_trace_list.append(longest_path_trace)
            longest_path_score_list.append(longest_path_score)

        # 三段階で initial_idx を決定 (maximum longest_path_score -> maximum query_start_end_gap -> maximum gap)
        # max のスコアを与える conserved_region のインデックスを取得
        longest_path_score = np.max(longest_path_score_list)
        initial_idx_list_with_max_score = np.where(longest_path_score_list == longest_path_score)[0]
        if len(initial_idx_list_with_max_score) == 1:   # この場合分けがないと、下記の np.diff でエラーが出る
            return longest_path_trace_list[initial_idx_list_with_max_score[0]], longest_path_score
        else:
            # スコア max のうち、query の start と end のギャップが最大のものを採用する (もっとも多くの場合は全部同じになるけど)
            query_start_end_gap_list = []   # __len__ = len(initial_idx_list_with_max_score)
            gap_max_list = []               # __len__ = len(initial_idx_list_with_max_score)
            for initial_idx in initial_idx_list_with_max_score:
                longest_path_trace = longest_path_trace_list[initial_idx]
                # longest_path_trace の数値は circular の切れ目である一箇所を除いて昇順で並んでいる ので、下記により gap の幅を求められる
                conserved_regions_ordered = conserved_regions.conserved_regions[longest_path_trace, :]
                query_start_idx = conserved_regions_ordered[:, 2].argmin()
                query_end_idx = conserved_regions_ordered[:, 3].argmax()
                query_start_end_gap_list.append((conserved_regions_ordered[query_start_idx, 0] - conserved_regions_ordered[query_end_idx, 1])%N_ref)
                # さらにその中で、query_start_gap 以外の gap が最大なものを選出する
                    # query_start_gap は np.diff の値がマイナスになるので自動的に除かれる
                    # 最初と最後の np.diff の値が計算結果に含まれないので別途追加する
                    # conserved_regions の長さが 1 の場合もあるので、その場合分けも処理する
                diff = np.diff(conserved_regions_ordered[:, 2:].flatten())[1::2]
                if len(diff) > 0:
                    gap_max_list.append(max(
                        diff.max(), 
                        conserved_regions_ordered[0, 2] - conserved_regions_ordered[-1, 3]
                    ))
                else:
                    gap_max_list.append(conserved_regions_ordered[0, 2] - conserved_regions_ordered[-1, 3])
            # query の start と end のギャップが最大ものも
            query_start_end_gap_max = max(query_start_end_gap_list)
            q_se_gap_max_idx_list = np.where(np.array(query_start_end_gap_list) == query_start_end_gap_max)[0]
            gap_max_idx = np.array(gap_max_list)[q_se_gap_max_idx_list].argmax()
            initial_idx = initial_idx_list_with_max_score[q_se_gap_max_idx_list][gap_max_idx]
            return longest_path_trace_list[initial_idx], longest_path_score

    @classmethod
    def search_longest_path(cls, conserved_regions_copied):  # conserved_regions.shape = (N, 4)  # ref_start, ref_end, query_start, query_end
        # conserved_regions = np.vstack((conserved_regions, [8000, 8100, 2000, 2100], [0, 80, 1000, 1100], [100, 400, 1300, 1500]))
        """
        2つの conserved_region (cr1, cr2) が cr1 -> cr2 の有向エッジをもつための条件
        - cr1_end_ref <= cr2_start_ref
        - cr1_end_query <= cr2_start_query
        - 上記 4 つの数値で囲われる長方形に完全に含まれるような conserved_region が無い
        """

        # # start >= end となっているものは除外する（circularな都合）
        # conserved_regions_copied.conserved_regions = conserved_regions_copied.conserved_regions[(np.diff(conserved_regions_copied.conserved_regions[:, :2], axis=1).flatten() > 0) * (np.diff(conserved_regions_copied.conserved_regions[:, 2:], axis=1).flatten() > 0), :]

        # 0, 0 から始まる conserved_region が必須！
        where_all_zero = np.where((conserved_regions_copied.conserved_regions[:, (0, 2)] == 0).all(axis=1))[0]
        assert len(where_all_zero) == 1
        initial_node_idx = where_all_zero[0]
        # コネクションマトリックスを生成
        connection_matrix_ref = cls.get_1d_connection(conserved_regions_copied.conserved_regions[:, :2])     # 以下で統合する
        connection_matrix_query = cls.get_1d_connection(conserved_regions_copied.conserved_regions[:, 2:])   # 以下で統合する
        connection_constraint = cls.get_connection_constraint(conserved_regions_copied.conserved_regions)    # 以下で統合する
        connection_matrix = ConnectionMatrix(
            (connection_matrix_ref + connection_matrix_query) * connection_constraint * np.diff(conserved_regions_copied.conserved_regions[:, :2], axis=1).flatten(),    # エッジの重みは、ノードの終点の conserved_region の長さとする（その後の解析で始点は共通で始めるので、こうすると最後に終点処理しいないでいい）。
            initial_node_idx=initial_node_idx, 
            initial_node_score=np.diff(conserved_regions_copied.conserved_regions[initial_node_idx, :2])[0]
        )
        # end < start  となっているものは除外する（circularな都合）
        connection_matrix.ignore_nodes(np.where(
            (np.diff(conserved_regions_copied.conserved_regions[:, :2], axis=1).flatten() < 0) + 
            (np.diff(conserved_regions_copied.conserved_regions[:, 2:], axis=1).flatten() < 0)
        )[0])
        connection_matrix.remove_unreachable_branches()
        longest_path_trace, longest_path_score = connection_matrix.execute_dijkstras_algorithm()

        # print(longest_path_trace)
        # cls.draw_connection(conserved_regions_copied.conserved_regions, connection_matrix.connection_matrix)
        # quit()

        return longest_path_trace, longest_path_score
    @classmethod
    def get_1d_connection(cls, conserved_1d_regions):    # conserved_1d_region.shape = (N, 2) # 2 means (start, end)
        conserved_1d_regions = np.copy(conserved_1d_regions).astype(float)

        # 以下が無いと、前の region の end と 次の region の start が同じだった場合、そのコネクトを許容するかどうかは region の登場順番に応じてランダムに決定されてしまう。
        # get_connection_constraint でどちらにしろフィルターされるからここでは無くても大丈夫か？と思いきや、どれとどれがコネクトされるかも変わってくるので必要
        conserved_1d_regions[:, 0] -= 0.1

        N_regions = conserved_1d_regions.shape[0]
        connection_matrix = np.zeros((N_regions, N_regions), dtype=bool)
        # cr1_end - cr2_start の間に他の cr があるかを検出
        start_end_argsort_idx = np.unravel_index(np.argsort(conserved_1d_regions, axis=None), (N_regions, 2))[0]
        """
        start_end_argsort_idx は同じ数字は 0-N_regions - 1 までの数が二回ずつ登場する一次元の配列 (start, end の idx)
        - コネクトの例
            1 1 85 85 (1->85)
            1 1 85 85 (1->85)
            16 16 28 25 28 25 (16->28, 16->25)
            26 29 26 27 29 27 23 23 (26->27, 29->23, 27->23)
        """
        N_of_appearance_list = np.zeros((N_regions), dtype=int) # 二回登場かつコネクションを持つ場合は、値を 3 とする：これ以上コネクションを持つ可能性が無い場合は -1 とする
        for idx in start_end_argsort_idx:
            # 初登場の場合：二回登場済みのもの（値が3も含まれる）とコネクションを形成
            if N_of_appearance_list[idx] == 0:
                # コネクション作成：すでに二回登場しているものとのみ、コネクションを作成
                edge_start_loc_list = N_of_appearance_list >= 2
                edge_start_idx_list = np.where(edge_start_loc_list)[0]
                connection_matrix[(edge_start_idx_list, [idx for i in edge_start_idx_list])] = 1

                # 後処理
                N_of_appearance_list[edge_start_loc_list] = 3
                N_of_appearance_list[idx] += 1
            # 二回目登場の場合：新たにコネクションが作られることはない
            else:
                # すでに二回登場して、コネクションを持ってるやつらを打ち切る
                N_of_appearance_list[N_of_appearance_list == 3] = -1
                # 後処理
                N_of_appearance_list[idx] += 1
        assert all(i in (-1, 2) for i in N_of_appearance_list)
        return connection_matrix
    @classmethod
    def get_connection_constraint(cls, conserved_regions):
        # ref, query, どちらにおいても, 次の region の start が前の region の end より大きくなっているもののみを選択
        constraint_ref = conserved_regions[:, 1][:, np.newaxis] < conserved_regions[:, 0]
        constraint_query = conserved_regions[:, 3][:, np.newaxis] < conserved_regions[:, 2]
        return constraint_ref * constraint_query
    @classmethod
    def draw_connection(cls, conserved_regions, connection_matrix):
        import matplotlib.pyplot as plt
        for i, (ref_start, ref_end, query_start, query_end) in enumerate(conserved_regions):
            plt.plot((ref_start, ref_end), (query_start, query_start), c="r", alpha=0.2, linestyle="--")
            plt.plot((ref_start, ref_end), (query_end, query_end), c="r", alpha=0.2, linestyle="--")
            plt.plot((ref_start, ref_start), (query_start, query_end), c="r", alpha=0.2, linestyle="--")
            plt.plot((ref_end, ref_end), (query_start, query_end), c="r", alpha=0.2, linestyle="--")
            plt.text(ref_start, query_end, f"{i}")
        for i, j in zip(*np.where(connection_matrix)):
            print(i, j)
            plt.plot(conserved_regions[([i, j], [1, 0])], conserved_regions[([i, j], [3, 2])])
        plt.xlabel("ref")
        plt.ylabel("query")
        plt.show()

class WindowAnalysis():
    consecutive_true_threshold = 20 # ref_seq (seq1) と query_seq (seq2) がこの値より大きい回数連続して同じ場合、保存領域の候補とする。
    @classmethod
    def exec(cls, seq1, seq2) -> None:  # they should be len(seq1) == len(seq2)
        a = (seq1 == seq2).astype(int)
        transition_list = a[1:] - a[:-1]
        true_start_idx_list = np.where(transition_list == 1)[0] + 1
        true_end_idx_list = np.where(transition_list == -1)[0]
        if a[0]:
            true_start_idx_list = np.hstack(([0], true_start_idx_list))
        if a[-1]:
            true_end_idx_list = np.hstack((true_end_idx_list, [len(a) - 1]))
        assert len(true_start_idx_list) == len(true_end_idx_list)
        # 最初と最後は繋げない (a[0] == a[-1] == True の場合でも、そのまま放置：まとめたりしないよ)
        
        # True が連続している長さを計算、threshold 以上の場所を取得する
        consecutive_true_length_list = true_end_idx_list - true_start_idx_list + 1
        over_threshold_where = consecutive_true_length_list > cls.consecutive_true_threshold
        return true_start_idx_list[over_threshold_where], true_end_idx_list[over_threshold_where]

        # import matplotlib.pyplot as plt
        # plt.plot(range(len(consecutive_true_length_list)), consecutive_true_length_list)
        # plt.show()
        # quit()






# -*- coding: utf-8 -*-

import re
import textwrap
import numpy as np
import scipy.spatial.distance as distance
from tqdm import tqdm
from typing import List
from pathlib import Path
from itertools import combinations
from scipy.cluster.hierarchy import linkage, leaves_list
from pulp import LpProblem, LpVariable, PULP_CBC_CMD, LpStatus, LpBinary, LpMinimize, lpSum
import matplotlib.pyplot as plt
from matplotlib import rc
from matplotlib.backends.backend_pdf import PdfPages
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

from ..modules import my_classes as mc
from ..modules import ref_query_alignment as rqa
rc('font', **{'family':'sans-serif','sans-serif':[mc.sans_serif_font_master]})

__all__ = ["execute_grouping"]

def execute_grouping(ref_seq_list: List[mc.MyRefSeq], param_dict: dict, save_dir: Path, ref_seq_aliases: list=None):
    assert (ref_seq_aliases is None) or (len(ref_seq_list) == len(ref_seq_aliases))
    recommended_grouping_path = save_dir.parent / RecommendedGrouping.file_name
    # load if any previous score_matrix
    skip = False
    if recommended_grouping_path.exists():
        recommended_grouping = RecommendedGrouping()
        recommended_grouping.load(recommended_grouping_path)
        if recommended_grouping.assert_data(ref_seq_list, param_dict):
            recommended_grouping.param_dict = param_dict
            print("calculation of distance matrix: SKIPPED")
            skip = True
    recommended_grouping_path = save_dir / RecommendedGrouping.file_name
    if not skip:
        print("calculating distance matrix...")
        recommended_grouping = RecommendedGrouping(ref_seq_list, param_dict, ref_seq_aliases)
        # 距離行列/linkage 計算
        recommended_grouping.calc_distance_linkage()
        recommended_grouping.save(recommended_grouping_path)
        print("calculation: DONE")
    #　コンビネーション計算実行
    # print("\n### distance matrix ###")
    # print(recommended_grouping.distance_matrix)
    # print("\n### linkage result ###")
    # print(recommended_grouping.linkage_result.astype(int))
    print("\ndetermining number of groups...")
    recommended_grouping.determine_number_of_groups()
    print(f"determination: DONE\n\n{recommended_grouping.message}")
    # print("\n### cluster list ###")
    # print(recommended_grouping.cluster_list)
    print("\nexecuting plasmid assignment...")
    recommended_grouping.execute_knapsac()
    print("\nassignment: DONE")
    return recommended_grouping

class RecommendedGrouping(mc.MyTextFormat, mc.MyHeader):
    grouping_algorithm_version = "ga_0.2.1"
    linkage_method = "single"   # 最短距離法: クラスタ間で最も近いデータの距離
    file_name = "recommended_grouping.txt"
    wrap_width = 100
    def __init__(self, ref_seq_list: List[mc.MyRefSeq]=None, param_dict: dict=None, ref_seq_aliases: list=None):
        super().__init__()
        self.header += (
            f"\ngrouping_algorithm_version: {self.grouping_algorithm_version}"
            f"\nassignment_solver_version: {PlasmidAssignmentSolver.assignment_solver_version}"
        )
        self.ref_seq_list = ref_seq_list
        self.param_dict = param_dict
        self.distance_matrix = None
        self.linkage_result = None
        self.adopted_number_of_groups = None
        # attributes below can be empty when saving
        self.message = ""
        self.cluster_list = []
        self.assignment_matrix = np.array([[]], dtype=int)
        # hidden attributes
        self._ref_seq_aliases = ref_seq_aliases
        self.recommended_grouping_from_txt = None
        self.keys = [
            ("header", "str"), 
            ("datetime", "str"), 
            ("ref_seq_paths", "listPath"), 
            ("ref_seq_aliases", "list"), 
            ("ref_seq_hash_list", "list"), 
            ("param_dict", "dict"), 
            ("distance_matrix", "ndarray"), 
            ("linkage_result", "ndarray"), 
            ("linkage_method", "str"), 
            ("message", "str"), 
            # ("cluster_list", "listlist"),         # assignment_matrix(_str) からに完全に含まれる情報
            ("assignment_matrix_txt", "str"),       # アサインされてないところを "*" で表現するために、string で保存
            ("recommended_grouping_txt", "str"),    # assignment_matrix_txt より後に読み込まれる必要がある（可能ならすべてすべて全ての属性の中で最後に、かな？）。
        ]
    @property
    def ref_seq_paths(self) -> List[Path]:
        return [ref_seq.path for ref_seq in self.ref_seq_list]
    @ref_seq_paths.setter
    def ref_seq_paths(self, ref_seq_paths: list):
        if self.ref_seq_list is not None:
            raise Exception("error: self.ref_seq_list is already assigned")
        else:
            try:
                self.ref_seq_list = [mc.MyRefSeq(ref_seq_path) for ref_seq_path in ref_seq_paths]
            except: # 存在しないファイルが一つでもある場合: (ほぼ) path 情報だけ保存するシュードインスタンスを作成
                self.ref_seq_list = [mc.MyRefSeq.Pseudo(ref_seq_path) for ref_seq_path in ref_seq_paths]
    @property
    def ref_seq_aliases(self):
        if (self._ref_seq_aliases is None) or (len(self._ref_seq_aliases) != self.N_plasmids):
            return [f"P{i + 1}" for i in range(self.N_plasmids)]
        else:
            return self._ref_seq_aliases
    @ref_seq_aliases.setter
    def ref_seq_aliases(self, ref_seq_aliases):
        self._ref_seq_aliases = ref_seq_aliases
    @property
    def ref_seq_hash_list(self) -> List[str]:
        return [ref_seq.my_hash for ref_seq in self.ref_seq_list]
    @property
    def N_plasmids(self) -> int:
        return len(self.ref_seq_list)
    @property
    def notice_beg(self):
        return f"{'=' * (self.wrap_width // 2 - 4)} NOTICE {'=' * (self.wrap_width // 2 - 4)}"
    @property
    def notice_end(self):
        return f"{'=' * self.wrap_width}"
    """
    # 上下関係: 
        - self.cluster_list
        - self.assignment_matrix
        - self.assignment_matrix_txt (self.assignment_matrix から動的生成)
        - self.recommended_grouping (self.assignemnt_matrix から動的生成)
        - self.recommended_grouping_txt (self.recommended_grouping から動的生成)
    # 保存されるのは
        - self.assignment_matrix_txt
        - self.recommended_grouping_txt
    # テキストファイルから読み込まれる場合
        - self.assignment_matrix_txt
        から
        - self.cluster_list
        を生成
        self.recommended_grouping_txt から一時的に recommended_grouping を生成する
        それが self.assignment_matrix に矛盾が発生する場合 (ユーザーが勝手に書き換えた場合位など) は self.recommended_grouping_txt の内容を優先する
    """
    @property
    def recommended_grouping(self):
        if self.recommended_grouping_from_txt is None:
            return [[idx for idx in assignment_row if idx != -1] for assignment_row in self.assignment_matrix.T]
        else:
            return self.recommended_grouping_from_txt
    @property
    def recommended_grouping_txt(self):
        grouping_str = ""
        for group_idx, plasmid_idx_list in enumerate(self.recommended_grouping):
            grouping_str += f"=== Group {group_idx + 1} ===\n"
            for idx in plasmid_idx_list:
                grouping_str += f"{self.ref_seq_aliases[idx]}\t{self.ref_seq_paths[idx].name}\n"
            grouping_str += "\n"
        return grouping_str
    @recommended_grouping_txt.setter
    def recommended_grouping_txt(self, recommended_grouping_txt):
        """
        "=== Group ([0-9]+) ===" と 改行 をもとにグループを detect する
        それをもとに recommended_grouping を作り、self.assignment_matrix と比較して矛盾がないかを検証する
        """
        group_match_list = re.findall(r"=== Group ([0-9]+) ===\n((?:.+\t.+\n)*)\n", recommended_grouping_txt.strip() + "\n\n")
        plasmid_match_re = re.compile(r"(.+)\t(.+)\n")
        recommended_grouping = []
        for gropu_idx, (group_number, match) in enumerate(group_match_list):
            assert gropu_idx == int(group_number) - 1
            group = []
            for ref_seq_alias, ref_seq_name in plasmid_match_re.findall(match):
                # パス違いでファイル名が同じ可能性も考慮して、、プラスミド名が一致するかを確認
                idx = self.ref_seq_aliases.index(ref_seq_alias)
                assert self.ref_seq_paths[idx].name == ref_seq_name
                group.append(idx)
            recommended_grouping.append(group)
        """
        # 検証
            recommended_grouping_txt をもとに作った recommended_grouping
        と
            assignment_matrix (テキストを読み込む時の由来は assignment_matrix_txt) から作った self.recommended_grouping
        の間にに矛盾がないかを確認
        """
        if len(recommended_grouping) == len(self.recommended_grouping) == 0:
            print(f"\033[91m\n{self.notice_beg}\n{textwrap.fill('`recommended_grouping` was not set either from `assignment_marix_txt` or `recommended_groupoing_txt`.', self.wrap_width)}\n{self.notice_end}\033[0m\n")
            self.adopted_number_of_groups = None
        elif not all(sorted(rg1) == sorted(rg2) for rg1, rg2 in zip(recommended_grouping, self.recommended_grouping)):
            self.recommended_grouping_from_txt = recommended_grouping
            print(f"\033[91m\n{self.notice_beg}\n{textwrap.fill('Conflicts were found between provided `assignment_marix_txt` and `recommended_groupoing_txt`. Constraints of `assignment_marix_txt` was ignored.', self.wrap_width)}\n{self.notice_end}\033[0m\n")
            self.adopted_number_of_groups = len(group_match_list)
        else:
            self.adopted_number_of_groups = self.assignment_matrix.shape[1]
    @property
    def assignment_matrix_txt(self):
        assignment_matrix_txt = "\n".join(["\t".join([str(idx) if idx != -1 else "." for idx in assignment]) for assignment in self.assignment_matrix])
        return assignment_matrix_txt
    @assignment_matrix_txt.setter
    def assignment_matrix_txt(self, assignment_matrix_txt):
        if len(assignment_matrix_txt) != 0:
            self.assignment_matrix = np.array([[int(idx) if idx != "." else -1 for idx in assignment.split("\t")] for assignment in assignment_matrix_txt.split("\n")])
            self.cluster_list = [[idx for idx in assignment if idx != -1] for assignment in self.assignment_matrix.T]
        else:
            self.assignment_matrix = np.array([[]], dtype=int)
    @property
    def adopted_distance_threshold(self):
        """
        self.adopted_number_of_groups は、self.assignment_matrix(_txt) および self.recommended_grouping(_txt) と関連してくるので、
        def recommended_grouping_txt 内で代入が行われる。
        self.adopted_distance_threshold については、あまり気にしなくても良さそうなので、単純に self.message から正規表現で取得する。
        """
        m = re.search(r"\nadopted `distance_threshold`\t([0-9]+)\n", self.message)
        if m is None:
            return None
        else:
            return int(m.group(1))
    def assert_data(self, ref_seq_list: List[mc.MyRefSeq], param_dict: dict):
        if len(ref_seq_list) != len(self.ref_seq_list):
            return False
        elif len(ref_seq_list) != len(self._ref_seq_aliases):
            return False
        elif all(self.param_dict[key] == param_dict[key] for key in ['gap_open_penalty', 'gap_extend_penalty', 'match_score', 'mismatch_score']):
            return False
        else:
            hash_list_new = [ref_seq.my_hash for ref_seq in ref_seq_list]
            if hash_list_new == self.ref_seq_hash_list:
                return True
            else:
                return False
    def calc_distance_linkage(self):
        self.set_distance_matrix()
        self.set_linkage()
    def set_distance_matrix(self):
        self.distance_matrix = np.zeros((self.N_plasmids, self.N_plasmids), dtype=int)
        ref_seq_idx_list = np.arange(self.N_plasmids)
        ref_seq_idx_pairs = combinations(ref_seq_idx_list, 2)
        for idx_1, idx_2 in tqdm(ref_seq_idx_pairs, ncols=100, mininterval=0.05, leave=True, bar_format='{l_bar}{bar}{r_bar}', total=self.N_plasmids * (self.N_plasmids - 1) // 2):
            ref_seq_1 = self.ref_seq_list[idx_1]
            ref_seq_2 = self.ref_seq_list[idx_2]
            self.distance_matrix[idx_1, idx_2] = self.distance_matrix[idx_2, idx_1] = self.calc_distance(ref_seq_1, ref_seq_2)
        assert (self.distance_matrix >= 0).all()
    def calc_distance(self, ref_seq_1: mc.MyRefSeq, ref_seq_2: mc.MyRefSeq):
        my_optimized_aligner = rqa.MyOptimizedAligner(ref_seq_1, self.param_dict)
        ref_seq_2 = mc.MySeq(ref_seq_2._seq)
        ref_seq_2_rc = ref_seq_2.reverse_complement()
        conserved_regions = my_optimized_aligner.calc_circular_conserved_region(ref_seq_2)
        conserved_regions_rc = my_optimized_aligner.calc_circular_conserved_region(ref_seq_2_rc)
        if conserved_regions is not None:
            my_result = my_optimized_aligner.execute_circular_alignment_using_conserved_regions(ref_seq_2, conserved_regions)
        else:
            my_result = rqa.MyResult()
        if conserved_regions_rc is not None:
            my_result_rc = my_optimized_aligner.execute_circular_alignment_using_conserved_regions(ref_seq_2_rc, conserved_regions_rc)
        else:
            my_result_rc = rqa.MyResult()
        # return distance
        if (my_result.score <= 0) and (my_result_rc.score <= 0):
            return len(ref_seq_1) + len(ref_seq_2)
        elif my_result.score >= my_result_rc.score:
            return my_result.calc_levenshtein_distance()
        else:
            return my_result_rc.calc_levenshtein_distance()
    def set_linkage(self):
        dArray = distance.squareform(self.distance_matrix)
        self.linkage_result = linkage(dArray, method=self.linkage_method).astype(int)   # distance が全て整数で、かつ linkage_method が average などでないため整数型にして ok
        # print(leaves_list(self.linkage_result))
        # print(self.linkage_result)
    def determine_number_of_groups(self):
        # 指定されたグループと distance_threshold の調整
        end_row_to_iter_plus, adopted_distance_threshold, self.adopted_number_of_groups = self.investigate_linkage_result()
        assert self.param_dict["distance_threshold"] <= adopted_distance_threshold
        assert self.param_dict["number_of_groups"] <= self.adopted_number_of_groups

        # メッセージ
        self.message = (
            f"provided `distance_threshold`\t{self.param_dict['distance_threshold']}\n"
            f"provided `number_of_groups`\t{self.param_dict['number_of_groups']}\n"
            f"adopted `distance_threshold`\t{adopted_distance_threshold}\n"
            f"adopted `number_of_groups`\t{self.adopted_number_of_groups}\n"
            f"`end_row_to_iter_plus`\t{end_row_to_iter_plus}"
        )
        # [CASE 5]
        if (self.param_dict["distance_threshold"] == adopted_distance_threshold) and (self.param_dict["number_of_groups"] == self.adopted_number_of_groups):
            core_message = ""
        # [CASE 2, 3, 6]
        elif (self.param_dict["distance_threshold"] < adopted_distance_threshold) and (self.param_dict["number_of_groups"] == self.adopted_number_of_groups):
            core_message = "\n" + textwrap.fill(
                f"Higher `distance_threshold` {adopted_distance_threshold} was used and the provided value {self.param_dict['distance_threshold']} was ignored."
                f" This is because the use of higer `distance_threshold` returns safer pre-survey results, and `distance_threshold` was able to be increased without changing the provided `number_of_groups` {self.param_dict['number_of_groups']}."
                , self.wrap_width
            )
        # [CASE 4]
        elif self.param_dict["number_of_groups"] < self.adopted_number_of_groups:
            core_message = (
                f"Provided `number_of_groups` {self.param_dict['number_of_groups']} is to small to meet the `distance_threshold` {self.param_dict['distance_threshold']}."
                f" The `number_of_groups` {self.adopted_number_of_groups} will be used instead."
            )
            if self.param_dict["distance_threshold"] > 1:
                core_message += f" To achieve number of groups {self.param_dict['number_of_groups']}, lower the `distance_threshold`."
            else:
                core_message += f" Number of groups lower than {self.param_dict['number_of_groups']} cannot be achieved as some plasmid maps share exactly the same sequences."
            if self.param_dict["distance_threshold"] < adopted_distance_threshold:
                core_message += (
                    f" Apart from that, higher `distance_threshold` {adopted_distance_threshold} was used and the provided value {self.param_dict['distance_threshold']} was ignored."
                    f" This is because the use of higer `distance_threshold` returns safer pre-survey results, "
                    f"and `distance_threshold` was able to be increased without changing the new `number_of_groups` {self.adopted_number_of_groups}."
                )
            core_message = "\n" + textwrap.fill(core_message, self.wrap_width)
        else:
            raise Exception(f"error:\n{self.message}\end_row_to_iter_plus:\t{end_row_to_iter_plus}")
        self.message = f"{self.notice_beg}\n{self.message}{core_message}\n{self.notice_end}"

        # クラスタ形成
        self.cluster_list = [[i] for i in range(self.N_plasmids)]
        for idx1, idx2, d, number_of_sub_cluster in self.linkage_result[:end_row_to_iter_plus, :]:
            self.cluster_list.append(self.cluster_list[int(idx1)] + self.cluster_list[int(idx2)])
            self.cluster_list[int(idx1)] = None
            self.cluster_list[int(idx2)] = None
        self.cluster_list = [c for c in self.cluster_list if c is not None]
    def investigate_linkage_result(self):
        """
        # about conditions
            distance_threshold は大きい方が良い (安全)
            number_of_groups は小さい方が良い (安い)
        # distance_threshold "未満" の距離のものは同じクラスターに含めなくてはいけないと定めている (前のバージョンのSAVEMONEYも同じ)。
            distance_threshold が 7 の時は、距離が 7 のプラスミドペアは、同じクラスタに含めなくともよい
            言い換えると、グループ内での最低距離が 7 "以上" になるようにするということ。
        # linkage results:
            [[idx1, idx2, dist, N_gr], 
             [idx1, idx2, dist, N_gr], 
             ...
             [idx1, idx2, dist, N_gr]]
        # Example of self.linkage_result (when distance_threshold=7, number_of_groups=3)
            実際に使われる数字は、 distance_threshold - 1 = 6 となる。
            この場合、distance <= 6 のペアは同じクラスタに入る -> グループ内の distance >=7 が保証される。
            row  dist  N_gr/max_number_of_sub_cluster [CASE 1]
            0      [1]   [2]
            1      [2]   [3]
            2      [5]   [3]
            3      [6]   [3]
            4      [6]    4 **** distance_threshold - 1 = 10 - 1, number_of_groups = max(4, 3)
            5      10     4
            row  dist  N_gr/max_number_of_sub_cluster [CASE 2]
            0      [1]   [2]
            1      [6]   [3]
            2      [6]   [3]
            3       7    [3] **** distance_threshold - 1 = 8 - 1, number_of_groups = max(3, 3)
            4       8     4
            5      10     4
            row  dist  N_gr/max_number_of_sub_cluster [CASE 3]
            0      [6]    2
            1      [6]    3
            2       7    [3]
            3       7    [3]
            4      10    [3] **** distance_threshold - 1 = 11 - 1, number_of_groups = max(3, 3)
            5      11     4
            row  dist  N_gr/max_number_of_sub_cluster [CASE 4]
            0      [1]   [2]
            1      [1]   [3]
            2      [1]   [3]
            3      [1]   [3]
            4      [6]    4
            5      [6]    4  この行で止めてしまうと、違うクラスタ内のプラスミドペアで distance_threshold を満たさないペアができる可能性がある
            6      [6]    5  **** distance_threshold = max(6+1, 7), number_of_groups = 5
            row  dist  N_gr/max_number_of_sub_cluster [CASE 5]
            0      [1]   [2]
            1      [1]   [3]
            2      [1]   [3]
            3      [1]   [3] **** distance_threshold - 1 = 7 - 1, number_of_groups = max(3, 3)
            4       7     4
            5       7     4
            6       7     5
            row  dist  N_gr/max_number_of_sub_cluster [CASE 6]
            0       9    [2] **** distance_threshold - 1 = 9 - 1, number_of_groups = max(2, 3)
            1       9     4
            2       9     4
        """
        # self.linkage_result を上の行から見て行って、その行までに登場した最大の number_of_sub_cluster のリストを作る
        assert self.linkage_result[0, 3] == 2
        assert self.linkage_result[-1, 3] == self.linkage_result[:, 3].max() == self.N_plasmids
        assert 0 < self.param_dict["distance_threshold"] #<= self.linkage_result[:, 2].max() + 1
        assert 0 < self.param_dict["number_of_groups"] <= self.N_plasmids
        # 事前準備
        max_n = 0
        max_number_of_sub_cluster = []
        for number_of_sub_cluster in self.linkage_result[:, 3]:
            if number_of_sub_cluster > max_n:
                max_n = number_of_sub_cluster
            max_number_of_sub_cluster.append(max_n)
        # 実行
        for end_of_iter_plus, (dist, number_of_sub_cluster) in enumerate(zip(self.linkage_result[:, 2], max_number_of_sub_cluster)):
            if (dist > self.param_dict["distance_threshold"] - 1) and (number_of_sub_cluster > self.param_dict["number_of_groups"]):
                if end_of_iter_plus == 0:
                    return end_of_iter_plus, dist, 1
                else:
                    return end_of_iter_plus, dist, max(max_number_of_sub_cluster[end_of_iter_plus-1], self.param_dict["number_of_groups"])
        return self.linkage_result.shape[0], max(dist+1, self.param_dict["distance_threshold"]), number_of_sub_cluster
    def execute_knapsac(self):
        cluster_size_list = [len(cluster) for cluster in self.cluster_list]
        assert max(cluster_size_list) <= self.adopted_number_of_groups
        assert sum(cluster_size_list) == self.distance_matrix.shape[0]

        """
        # 目的関数までまとめて一気にゴリ押ししてsolveできるのでは?
        # 目的関数: idx_pairs_to_be_separated の組み合わせで、採用されなかったものの idx が最大になるようにすること
        # 時間がかかりすぎるようなら、やはり二段階に分離して行うべきか（まずプラスミドのクラスタ - グループの位置を決めて、そのあとプラスミドを配置していく）
        """
        # 許される組み合わせのペアの中で、最も距離が近いものを昇順に取得する
        inf_value = self.distance_matrix.max() + 1
        distance_matrix = self.distance_matrix * np.tri(*self.distance_matrix.shape, k=0, dtype=int) + np.tri(*self.distance_matrix.shape, k=0, dtype=int).T * inf_value    # 対角線より上の部分を -1 にする
        for cluster in self.cluster_list:   # 同じクラスタに含まれるプラスミドのペアは、ここでは考慮しない（LP の制約条件で記述する）
            for idx1, idx2 in combinations(cluster, 2):
                distance_matrix[idx1, idx2] = distance_matrix[idx2, idx1] = inf_value
        unique, inverse = np.unique(distance_matrix, return_inverse=True)   # 距離が小さいものから順に取得
        inverse = inverse.reshape(distance_matrix.shape)
        idx_pairs_in_order_of_distance = []                                          # ペアの idx を取得
        for i, distance in enumerate(unique[:-1]):
            idx_pairs_in_order_of_distance.extend([PlasmidAssignmentSolver.sort_idx(idx1, idx2) for idx1, idx2 in zip(*np.where(inverse == i))])

        # """
        # # まず、クラスタ内のプラスミドの区別なく、組分けを行う。これにより、グループ間のプラスミドの個数がバラつくことを防ぐ
        # # その後、距離が短いものを最優先に、異なるクラスタに分類していく
        # """

        # 実行
        pas = PlasmidAssignmentSolver(self.adopted_number_of_groups, self.cluster_list, idx_pairs_in_order_of_distance) # 一気に解くのは線形じゃないからダメそう
        self.assignment_matrix = pas.execute()
    def load(self, load_path):
        ver_tmp = self.grouping_algorithm_version
        linkage_method_tmp = self.linkage_method
        super().load(load_path)
        if ver_tmp != self.get_version("grouping_algorithm_version"):
            print(
                f"\033[91m\n{self.notice_beg}\nmismatch of file version was found in `{load_path}`"
                f"\nexpected file version:\t{ver_tmp}\ndetected file version:\t{self.get_version('grouping_algorithm_version')}\n{self.notice_end}\033[0m"
            )
        if linkage_method_tmp != self.linkage_method:
            print(
                f"\033[91m\n{self.notice_beg}\nmismatch of `linkage_method` was found in `{load_path}`"
                f"\nexpected `linkage_method`:\t{linkage_method_tmp}\ndetected `linkage_method`:\t{self.linkage_method}\n{self.notice_end}\033[0m"
            )
        if PlasmidAssignmentSolver.assignment_solver_version != self.get_version("assignment_solver_version"):
            print(
                f"\033[91m\n{self.notice_beg}\nmismatch of solver version was found in `{load_path}`"
                f"\nexpected solver version:\t{ver_tmp}\ndetected solver version:\t{self.get_version('assignment_solver_version')}\n{self.notice_end}\033[0m"
            )
        self.path = load_path
    def draw_heatmaps(self, save_dir=None, display_plot=None):
        # 複数 PDF 作成準備
        if save_dir is not None:
            pdf = PdfPages((Path(save_dir) / self.file_name).with_suffix(".pdf"))

        # 描画
        for group_idx, group in enumerate(self.recommended_grouping):
            self.draw_heatmap(group_idx, group)
            # 保存? 表示? そのまま閉じる?
            if save_dir is not None:
                pdf.savefig()
            if display_plot is None:
                return
            elif display_plot:
                plt.show()
            else:
                plt.clf()
                plt.close()

        # 複数 PDF 作成事後処理
        if save_dir is not None:
            pdf.close()
    def draw_heatmap(self, group_idx, plasmid_idx_list_to_highlight):
        # font style params
        max_alias_width = max(len(ref_seq_alias) for ref_seq_alias in self.ref_seq_aliases)
        y_tick_labels = [f"{ref_seq_alias:<{max_alias_width}} {ref_seq_path.name}" for ref_seq_alias, ref_seq_path in zip(self.ref_seq_aliases, self.ref_seq_paths)]
        tick_font_size = 10
        # fig size parmas (fixed margin with absoute values, not percentile)
        heatmap_cell_unit = 0.5
        top_margin_unit = 0.6
        left_margin_unit = 1.0
        bottom_margin_unit = 0.8
        right_margin_unit = 0.5 + tick_font_size/100 * 0.9 * max(len(y_tick_laabel) for y_tick_laabel in y_tick_labels)
        bar_width_percentile = 3            # color bar のサイズは、heatmap サイズに応じて決める
        bbox_to_anchor_x0_percentile = -5   # color bar のサイズは、heatmap サイズに応じて決める
        fig_width = heatmap_cell_unit * self.N_plasmids + left_margin_unit + right_margin_unit
        fig_height = heatmap_cell_unit * self.N_plasmids + top_margin_unit + bottom_margin_unit
        # color styel params
        cbar_max = 30   #score_matrix.max()#
        
        # make matplotlib fig
        fig, ax= plt.subplots(1, 1, figsize=(fig_width, fig_height))
        im = ax.imshow(self.distance_matrix, cmap="YlGn", vmin=0, vmax=cbar_max)
        axins = inset_axes(ax, width=f"{bar_width_percentile}%", height="100%", loc='lower left', bbox_to_anchor=(bbox_to_anchor_x0_percentile/100, 0., 1, 1), bbox_transform=ax.transAxes, borderpad=0)
        bar = plt.colorbar(im, cax=axins)
        axins.yaxis.set_ticks_position('left')

        # Loop over data dimensions and create text annotations.
        for i in range(self.N_plasmids):
            for j in range(self.N_plasmids):
                value = self.distance_matrix[i, j]
                if value > 9999:
                    value_font_size = 8
                else:
                    value_font_size = 10
                if value < cbar_max / 2:
                    text = ax.text(j, i, str(value), ha="center", va="center", color="k", fontsize=value_font_size)
                else:
                    text = ax.text(j, i, str(value), ha="center", va="center", color="w", fontsize=value_font_size)

        # Show all ticks and label them with the respective list entries
        # x
        if max_alias_width > 5:
            x_tick_font_size = tick_font_size * 5 / max_alias_width
        else:
            x_tick_font_size = tick_font_size
        ax.xaxis.set_ticks_position("top")
        ax.set_xticks(np.arange(self.N_plasmids))
        ax.set_xticklabels(labels=[i for i in self.ref_seq_aliases], fontsize=x_tick_font_size, fontname="monospace")

        # y
        ax.yaxis.set_ticks_position("right")
        ax.set_yticks(np.arange(self.N_plasmids))
        ax.set_yticklabels(labels=y_tick_labels, fontsize=tick_font_size, fontname="monospace")

        # highlight heatmap cells
        for i in plasmid_idx_list_to_highlight:
            for j in plasmid_idx_list_to_highlight:
                if i == j:
                    continue
                else:
                    rect = plt.Rectangle((i-0.45, j-0.45), 0.9, 0.9, fill=False, color="r")
                    ax.add_patch(rect)
        # highlight tick labels
        for idx, xtick in enumerate(ax.get_xticklabels()):
            if idx in plasmid_idx_list_to_highlight:
                xtick.set_color("r")
        for idx, ytick in enumerate(ax.get_yticklabels()):
            if idx in plasmid_idx_list_to_highlight:
                ytick.set_color("r")

        ax.text(
            x=-0.5, y=self.N_plasmids, s=
            f'Group {group_idx+1}: {{{", ".join([alias for i, alias in enumerate(self.ref_seq_aliases) if i in plasmid_idx_list_to_highlight])}}}', 
            weight='bold', va="bottom"
        )
        ax.text(
            x=-0.5, y=self.N_plasmids+0.1, s=
            f'distance_threshold: {self.adopted_distance_threshold or "?"}\n'
            f'number_of_groups  : {self.adopted_number_of_groups or "?"}', 
            fontsize=tick_font_size, fontname="monospace", va="top"
        )
        plt.subplots_adjust(left=left_margin_unit/fig_width, bottom=bottom_margin_unit/fig_height, right=1-right_margin_unit/fig_width, top=1-top_margin_unit/fig_height)
    def get_ref_seq_path_list_in_group(self, group_idx):
        return [self.ref_seq_paths[idx] for idx in self.recommended_grouping[group_idx]]

class PlasmidAssignmentSolver():
    assignment_solver_version = "as_0.1.0"
    def __init__(self, adopted_number_of_groups, cluster_list, idx_pairs_in_order_of_distance) -> None:
        self.N_groups = adopted_number_of_groups
        self.cluster_list = cluster_list
        self.idx_pairs_in_order_of_distance = idx_pairs_in_order_of_distance    #　近い順
        self.idx_pairs_removed = []
        self.N_plasmids = sum(len(cluster) for cluster in self.cluster_list)
        self.N_clusters = len(self.cluster_list)
        self.N_plasmids_in_group_list = [self.N_plasmids // self.N_groups + 1 if i < self.N_plasmids%self.N_groups else self.N_plasmids // self.N_groups for i in range(self.N_groups)]
        assert (self.N_plasmids == sum(self.N_plasmids_in_group_list)) and (max(self.N_plasmids_in_group_list) - min(self.N_plasmids_in_group_list) <= 1)
        # 辞書的に使用する
        self.cluster_idx_from_plasmid_idx = np.empty(self.N_plasmids, dtype=int)
        for cluster_idx, cluster in enumerate(self.cluster_list):
            self.cluster_idx_from_plasmid_idx[cluster] = cluster_idx
    @staticmethod
    def sort_idx(idx1, idx2):
        if idx1 < idx2:
            return (idx1, idx2)
        elif idx1 > idx2:
            return (idx2, idx1)
        else:
            raise Exception(f"error: {idx1, idx2}")
    # 実行関数
    def execute(self):
        assignment_example_not_None = self.generate_example()
        pbar = tqdm(self.idx_pairs_in_order_of_distance, ncols=100, mininterval=0.02, leave=True, bar_format='{l_bar}{bar}{r_bar}')#, total=self.idx_pairs_in_order_of_distance)
        str_width = 4 + len(str(self.N_plasmids - 1)) * 2
        for idx_pair in pbar:
            pbar.set_postfix_str(f"examining edge {str(idx_pair):{str_width}}")
            self.idx_pairs_removed.append(idx_pair)
            assignment_example = self.generate_example()
            if assignment_example is None:
                del self.idx_pairs_removed[-1]
            else:
                assignment_example_not_None = assignment_example
        assert assignment_example_not_None is not None
        return assignment_example_not_None
    def generate_example(self):
        ########
        # 問題 #
        ########
        problem = LpProblem("GraphAssertion")
        ########
        # 変数 #
        ########
        # variable_box = (N_clusters, N_groups, N_plasmids)
        variable_box = [
            [
                [
                    LpVariable(f"c{cluster_idx}g{group_idx}p{plasmid_idx}", cat=LpBinary) for plasmid_idx in range(self.N_plasmids)
                ] for group_idx in range(self.N_groups)
            ] for cluster_idx in range(self.N_clusters)
        ]
        ###########
        # 制約条件 #
        ###########
        # クラスタ -- プラスミド 面からみた条件
        for cluster_idx, cluster in enumerate(self.cluster_list):
            for plasmid_idx in range(self.N_plasmids):
                if plasmid_idx in cluster:
                    problem += lpSum(variable_box[cluster_idx][group_idx][plasmid_idx] for group_idx in range(self.N_groups)) == 1
                else:
                    problem += lpSum(variable_box[cluster_idx][group_idx][plasmid_idx] for group_idx in range(self.N_groups)) == 0
        # クラスタ -- グループ 面からみた条件
        for cluster_idx in range(self.N_clusters):
            for group_idx in range(self.N_groups):
                problem += lpSum(variable_box[cluster_idx][group_idx][plasmid_idx] for plasmid_idx in range(self.N_plasmids)) <= 1
        # グループ の条件
        for group_idx, N_plasmids_in_group in enumerate(self.N_plasmids_in_group_list):
            problem += lpSum(variable_box[cluster_idx][group_idx][plasmid_idx] for cluster_idx in range(self.N_clusters) for plasmid_idx in range(self.N_plasmids)) == N_plasmids_in_group
        # 使ってはいけないペアの条件
        for plasmid_idx1, plasmid_idx2 in self.idx_pairs_removed:
            cluster_idx1 = self.cluster_idx_from_plasmid_idx[plasmid_idx1]
            cluster_idx2 = self.cluster_idx_from_plasmid_idx[plasmid_idx2]
            for group_idx in range(self.N_groups):
                problem += lpSum([variable_box[cluster_idx1][group_idx][plasmid_idx1], variable_box[cluster_idx2][group_idx][plasmid_idx2]]) <= 1
        ########
        # 求解 #
        ########
        solver = PULP_CBC_CMD(msg = False)
        result = problem.solve(solver=solver)
        status = LpStatus[result]
        if status == "Optimal":
            assignment_example = np.empty((self.N_clusters, self.N_groups), dtype=int)
            for cluster_idx in range(self.N_clusters):
                for group_idx in range(self.N_groups):
                    for plasmid_idx in range(self.N_plasmids):
                        if variable_box[cluster_idx][group_idx][plasmid_idx].value():
                            assignment_example[cluster_idx, group_idx] = plasmid_idx
                            break
                    else:
                        assignment_example[cluster_idx, group_idx] = -1
            return assignment_example
        else:
            return None

### これ単体で使用するには少し不完全だけど、めちゃめちゃプラスミド数が大きくなった時に何かの役に立つかもしれない ###
"""
一応、実行するには下記。
plasmid_loc_in_cluster_group_candidates = PlasmidAssignmentSolver.solve_simple_picross(distance_matrix.shape[0], adopted_number_of_groups, cluster_size_list)
for plasmid_loc_in_cluster_group in plasmid_loc_in_cluster_group_candidates:
    pa = PlasmidAssignment(plasmid_loc_in_cluster_group, self.cluster_list, unfavorable_idx_pairs)
    pa.execute()
"""
class PlasmidAssignment():
    def __init__(self, plasmid_loc_in_cluster_group, cluster_list, unfavorable_idx_pairs) -> None:
        assert plasmid_loc_in_cluster_group.dtype == int
        self.plasmid_loc_in_cluster_group = plasmid_loc_in_cluster_group
        self.cluster_list = cluster_list
        self.unfavorable_idx_pairs = unfavorable_idx_pairs
        self.unfavorable_idx_pairs_kept = []
        self.unfavorable_idx_pairs_removed = []
        self.cluster_size_list = np.array([len(cluster) for cluster in self.cluster_list])
        self.N_plasmids = sum(self.cluster_size_list)
        self.N_group = self.plasmid_loc_in_cluster_group.shape[1]
        # 辞書的に使用する
        self.cluster_idx_from_plasmid_idx = np.empty(self.N_plasmids, dtype=int)
        for cluster_idx, cluster in enumerate(self.cluster_list):
            self.cluster_idx_from_plasmid_idx[cluster] = cluster_idx
        # ややこしい
        """
        # pattern_list_for_each_plasmid example: 同じグループに所属しうるやつらが他のグループにあと何個残っているかを示すマトリックス。本関数内でアップデートされてく。
            cluster_list = [[0], [2], [5], [6], [11], [4, 7], [8, 13], [3, 12], [10, 1, 9]]
            where_plasmid = (shape = (len(cluster_list), adopted_number_of_groups))
                [[0 0 1]
                 [0 1 0]
                 [1 0 0]
                 [1 0 0]
                 [1 0 0]
                 [1 1 0]
                 [0 1 1]
                 [0 1 1]
                 [1 1 1]]

            [[-1  1  1  1  1  1  1  1  1  1  1  1  1  1]
             [ 1  1 -1  1  1  1  1  1  1  1  1  1  1  1]
             [ 1  1  1  1  1 -1  1  1  1  1  1  1  1  1]
             [ 1  1  1  1  1  1 -1  1  1  1  1  1  1  1]
             [ 1  1  1  1  1  1  1  1  1  1  1 -1  1  1]
             [ 2  2  2  2 -1  2  2 -1  2  2  2  2  2  2]
             [ 2  2  2  2  2  2  2  2 -1  2  2  2  2 -1]
             [ 2  2  2 -1  2  2  2  2  2  2  2  2 -1  2]
             [ 3 -1  3  3  3  3  3  3  3 -1 -1  3  3  3]]

            row: cluster_idx \ col: plasmid_idx
            [[ -1   1   1   1   1   1   1   1   1   1   1   1   1   1 ]  <- 基本的には、len(cluster_list[0]) の値が初期値として入る
             [  1   1  -1   1   1   1   1   1   1   1   1   1   1   1 ]  <- 基本的には、len(cluster_list[1]) の値が初期値として入る
             [  1   1   1   1   1  -1   1   1   1   1   1   1   1   1 ]  <- 基本的には、len(cluster_list[2]) の値が初期値として入る
             [  1   1   1   1   1   1  -1   1   1   1   1   1   1   1 ]  <- 基本的には、len(cluster_list[3]) の値が初期値として入る
             [  1   1   1   1   1   1   1   1   1   1   1   1  -1   1 ]  <- 基本的には、len(cluster_list[4]) の値が初期値として入る
             [  2   2   2   2 *-1*  2   2 *-1*  2   2   2   2   2   2 ]  <- 基本的には、len(cluster_list[5]) の値が初期値として入る ただし、[5(=cluster_idx), 4(=plasmid_idx)] と [5, 7] の場所の値は 1 となり、それは更新されない。
             [  2   2   2   2   2   2   2   2  -1   2   2   2   2  -1 ]
             [  2   2   2  -1   2   2   2   2   2   2   2   2  -1   2 ]
             [  3  -1   3   3   3   3   3   3   3  -1  -1   3   3   3 ]]
        # unfavorable_idx_pairs を iter して、pattern_list_for_each_plasmid をアップデートしていく。
            un  f avo rab le_i dx_ pai rs[ 0]  = ( 8,  3)  の処理 後を下 記に 示す。
            [[ -1   1   1   1   1   1   1   1   1   1   1   1   1   1 ]
             [  1   1  -1   1   1   1   1   1   1   1   1   1   1   1 ]
             [  1   1   1   1   1  -1   1   1   1   1   1   1   1   1 ]
             [  1   1   1   1   1   1  -1   1   1   1   1   1   1   1 ]
             [  1   1   1   1   1   1   1   1   1   1   1   1  -1   1 ]
             [  2   2   2   2  -1   2   2  -1   2   2   2   2   2   2 ]
             [  2   2   2  *1*  2   2   2   2  -1   2   2   2   2  -1 ] <- plasmid_idx = 3-8 はコネクトできなくなるので、[6(=cluster_idx where 8 is in), 3 (=plasmid_idx)] の位置と
             [  2   2   2  -1   2   2   2   2  *1*  2   2   2  -1   2 ] <- [7(=cluster_idx where 3 is in), 8 (=plasmid_idx)] の位置との値を 1 減らす。
             [  3  -1   3   3   3   3   3   3   3  -1  -1   3   3   3 ]]
        # さらに、unfavorable_idx_pairs[1] = (13, 3) の処理後を下記に示す。
            [[ -1   1   1   1   1   1   1   1   1   1   1   1   1   1  ]
             [  1   1  -1   1   1   1   1   1   1   1   1   1   1   1  ]
             [  1   1   1   1   1  -1   1   1   1   1   1   1   1   1  ]
             [  1   1   1   1   1   1  -1   1   1   1   1   1   1   1  ]
             [  1   1   1   1   1   1   1   1   1   1   1   1  -1   1  ]
             [  2   2   2   2  -1   2   2  -1   2   2   2   2   2   2  ]
             [  2   2   2  *0*  2   2   2   2  -1   2   2   2   2  -1  ] <- [6(=cluster_idx where 13 is in), 3 (=plasmid_idx)] の位置と
             [  2   2   2  -1   2   2   2   2   1   2   2   2  -1  *1* ] <- [7(=cluster_idx where 3 is in), 13 (=plasmid_idx)] の位置との値を 1 減らす。
             [  3  -1   3   3   3   3   3   3   3  -1  -1   3   3   3  ]]
        # この時点で、上のマトリックスの 3 列目 (Aとする) を満たせるような where_plasmid の列 (Bとする) が存在しない！計算式はちょっとややこしいので直下の実際の関数を参照してくれ
        # なので unfavorable_idx_pairs[1] = (13, 3) はスキップして、次へ進む
        """
        self.pattern_list_for_each_plasmid = np.ones((len(self.cluster_list), self.N_plasmids), dtype=int) * self.cluster_size_list[:, np.newaxis]
        for cluster_idx, cluster in enumerate(self.cluster_list):
            for plasmid_idx in cluster:
                self.pattern_list_for_each_plasmid[cluster_idx, plasmid_idx] = -1
        # ややこしい
        """ EXAMPLE: shape = (self.N_group, self.N_plasmids)
        # 列に True が 1 つしか出現しなくなったら、その plasmid の所属グループが決定したということ。
        [[False  True False False  True  True  True  True False  True  True  True False False]
         [False  True  True  True  True False False  True  True  True  True False  True  True]
         [ True  True False  True False False False False  True  True  True False  True  True]]
        """
        self.available_group_for_plasmid = np.zeros((self.N_group, self.N_plasmids), dtype=bool)  # bool
        for cluster_idx, cluster in enumerate(self.cluster_list):
            for plasmid_idx in cluster:
                for group_idx, plasmid_exists in enumerate(self.plasmid_loc_in_cluster_group[cluster_idx, :]):
                    if plasmid_exists:
                        self.available_group_for_plasmid[group_idx, plasmid_idx] = 1
        assert self.available_group_for_plasmid.any(axis=0).all()

        print("###")
        np.set_printoptions(linewidth=300)
        print(self.cluster_list)
        print(self.cluster_idx_from_plasmid_idx)
        print(self.plasmid_loc_in_cluster_group)
        print(self.unfavorable_idx_pairs)
        print(self.available_group_for_plasmid.astype(int))

        # 
        self.available_group_for_plasmid_initiated()


        # 使わないかも
        # adjacency_matrix = np.zeros((N_plasmids, N_plasmids), dtype=int)  # bool
        # for idx1, idx2 in unfavorable_idx_pairs:
        #     adjacency_matrix[idx1, idx2] = 1
        #     adjacency_matrix[idx2, idx1] = 1
    @staticmethod
    def solve_simple_picross(N_plasmids, adopted_number_of_groups, cluster_size_list):
        """
        # 配置
            縦 (axis=0) にクラスタ
            横 (axis=1) にグループ
        # 制約
            row の合計が cluster_size_list
            col の合計が N_plasmids_in_group_list
        # EXAMPLE
            [[0,  0,  1]    cluster_1 (size=1)
             [1,  0,  0]    cluster_2 (size=1)
             [0,  1,  0]    cluster_3 (size=1)
             [1,  0,  1]    cluster_4 (size=2)
             [1,  1,  1]]   cluster_5 (size=3)
              g1, g2, g3
              3   2   3 <- グループ間のプラスミドの個数の差は 1 以下になるようにする
        """
        N_plasmids_in_group_list = [N_plasmids // adopted_number_of_groups + 1 if i < N_plasmids%adopted_number_of_groups else N_plasmids // adopted_number_of_groups for i in range(adopted_number_of_groups)]
        assert (N_plasmids == sum(N_plasmids_in_group_list)) and (max(N_plasmids_in_group_list) - min(N_plasmids_in_group_list) <= 1)

        row_constraint = np.array(cluster_size_list)
        col_constraint = np.array(N_plasmids_in_group_list)
        N_row = len(row_constraint)
        N_col = len(col_constraint)

        # 問題
        problem = LpProblem("SimplePicross")
        # 変数（どこを 1 にするか）
        plasmid_loc_in_cluster_group = [
            [LpVariable(f"r{r}c{c}", cat=LpBinary) for c in range(N_col)]
            for r in range(N_row)
        ]
        # 制約条件の追加（目的関数はない：すべて数え上げる）
        for r in range(N_row):
            problem += lpSum(plasmid_loc_in_cluster_group[r][c] for c in range(N_col)) == row_constraint[r]
        for c in range(N_col):
            problem += lpSum(plasmid_loc_in_cluster_group[r][c] for r in range(N_row)) == col_constraint[c]
        # 求解
        solver = PULP_CBC_CMD()
        result = problem.solve(solver=solver)
        status = LpStatus[result]
        if status == "Optimal":
            result_array = np.empty((N_row, N_col), dtype=int)
            for i in range(N_row):
                for j in range(N_col):
                    result_array[i, j] = plasmid_loc_in_cluster_group[i][j].value()
        else:
            print("解なし")

        plasmid_loc_in_cluster_group_candidates = []  # 解を書き出す予定: TODO
        plasmid_loc_in_cluster_group_candidates.append(result_array)
        return plasmid_loc_in_cluster_group_candidates
    @staticmethod
    def __knapsac_core(adopted_number_of_groups, cluster_list, unfavorable_idx_pairs):
        N_plasmids = sum(len(cluster) for cluster in cluster_list)
        N_cluster = len(cluster_list)
        cluster_size_list = [len(cluster) for cluster in cluster_list]
        max_clusster_size = max(cluster_size_list)
        N_plasmids_in_group_list = [N_plasmids // adopted_number_of_groups + 1 if i < N_plasmids%adopted_number_of_groups else N_plasmids // adopted_number_of_groups for i in range(adopted_number_of_groups)]
        assert (N_plasmids == sum(N_plasmids_in_group_list)) and (max(N_plasmids_in_group_list) - min(N_plasmids_in_group_list) <= 1)

        ########
        # 問題 #
        ########
        problem = LpProblem("SimplePicross", LpMinimize)

        ########
        # 変数 #
        ########
        # どこにどのプラスミドを入れるかのインデックスを格納した3次元配列
        # where_plasmid[cluster_idx].shape = (adopted_number_of_groups, cluster_size_list[i])
        where_plasmid = []
        for cluster_idx, cluster_size in enumerate(cluster_size_list):
            group_and_in_cluster = []
            for group_idx in range(adopted_number_of_groups):
                in_cluster = []
                for idx_in_cluster in range(max_clusster_size):  # テンソルにするために、敢えて余計に入れる
                    var = LpVariable(f"c{cluster_idx}g{group_idx}i{idx_in_cluster}", cat=LpBinary)
                    in_cluster.append(var)
                    if idx_in_cluster > cluster_size - 1:
                        problem += var == 0
                group_and_in_cluster.append(in_cluster)
            where_plasmid.append(group_and_in_cluster)

        ################
        # 制約条件の追加 #
        ################
        # group に含まれるプラスミドの個数
        for group_idx in range(adopted_number_of_groups):
            problem += lpSum(
                where_plasmid[cluster_idx][group_idx][idx_in_cluster]
                for cluster_idx, cluster_size in enumerate(cluster_size_list)
                for idx_in_cluster in range(cluster_size)
            ) == N_plasmids_in_group_list[group_idx]
        # idx_in_cluster に含まれるプラスミドの個数
        for cluster_idx, cluster_size in enumerate(cluster_size_list):
            for idx_in_cluster in range(cluster_size):
                problem += lpSum(where_plasmid[cluster_idx][group_idx][idx_in_cluster] for group_idx in range(adopted_number_of_groups)) == 1
        # 同じグループに同一のクラスタ由来のプラスミドがが含まれないようにする
        for cluster_idx, cluster_size in enumerate(cluster_size_list):
            for group_idx in range(adopted_number_of_groups):
                problem += lpSum(where_plasmid[cluster_idx][group_idx][idx_in_cluster] for idx_in_cluster in range(cluster_size)) <= 1

        # 目的（最小化）関数：unfavorable_idx_pairs のスコアを、2**(max_idx - idx) などとしたら、最小化問題に帰着できる気がする
        # where_plasmid を、テンソルにする必要がある（0 を加えてクラスターサイズを揃えてある）
        score_tensor = np.arange()
        print()
        print(cluster_size_list)
        print(where_plasmid)
        problem += lpSum(

        )

        print(problem)
        quit()


        # 求解
        solver = PULP_CBC_CMD()
        result = problem.solve(solver=solver)
        status = LpStatus[result]
        if status == "Optimal":
            result_array = np.empty((N_cluster, adopted_number_of_groups), dtype=int)
            for cluster_idx, cluster_size in enumerate(cluster_size_list):
                for group_idx in range(adopted_number_of_groups):
                    idx_in_cluster_with_plasmid = -1
                    for idx_in_cluster in range(cluster_size):
                        i = where_plasmid[cluster_idx][group_idx][idx_in_cluster].value()
                        print(f"c{cluster_idx}g{group_idx}i{idx_in_cluster}: {int(i)}")
                        if i != 0:
                            assert idx_in_cluster_with_plasmid == -1
                            idx_in_cluster_with_plasmid = idx_in_cluster
                    result_array[cluster_idx, group_idx] = idx_in_cluster_with_plasmid
            print(result_array)
        else:
            print("解なし")


        quit()
    ##################
    # MAIN FUNCTIONS #
    ##################
    def available_group_for_plasmid_initiated(self):
        # pattern_list_for_each_plasmid/unfavorable_idx_pairs の処理
        for plasmid_idx1, available_group in enumerate(self.available_group_for_plasmid.T):
            for plasmid_idx2, allowed in enumerate(self.available_group_for_plasmid[available_group, :].sum(axis=0)):
                if (not allowed) and (plasmid_idx1 < plasmid_idx2):
                    self.remove_plasmid_connection_core(plasmid_idx1, plasmid_idx2)
        self.process_lonly_connection()
    # 実行関数
    def execute(self):
        """
        1. エッジを距離の短い順に除去する (self.remove_plasmid_connection)
            1-1. 除去すると矛盾が生じてしまう場合、エッジの除去を取りやめて次のループへ
            1-2. 除去 "できそう" な場合 (できるとはまだ言っていいない)
                1-2-0. 以下は、必要十分なエッジの除去を行う操作であり、恣意性は入らない（はず）
                1-2-1. 
        """
        print("EXECUTE")
        for idx1, idx2 in self.unfavorable_idx_pairs:
            print("IDX", idx1, idx2)
            if (idx1, idx2) in self.unfavorable_idx_pairs_kept:
                continue
            elif (idx1, idx2) in self.unfavorable_idx_pairs_removed:
                continue
            processed = self.remove_plasmid_connection(idx1, idx2)
            if processed:
                continue
            else:
                continue
        # チェック
        self.assert_available_group_for_plasmid()
        print(self.pattern_list_for_each_plasmid)
        print(self.available_group_for_plasmid)
        print(len(self.unfavorable_idx_pairs))
        print(len(self.unfavorable_idx_pairs_kept))
        print(len(self.unfavorable_idx_pairs_removed))
        print(self.unfavorable_idx_pairs)
        print(self.unfavorable_idx_pairs_kept)
        print(self.unfavorable_idx_pairs_removed)
        quit()
    def generate_available_group(self, idx):
        # self.available_group_for_plasmid はまだアプデされない
        # self.pattern_list_for_each_plasmid[:, idx] について、グループに所属できるかかどうかを　False, True のリスト (len=self.N_group) で返す。
        available_group = ((1 - self.pattern_list_for_each_plasmid[:, idx:idx+1] + self.plasmid_loc_in_cluster_group) != 2).all(axis=0)
        """
        available_group が適応された場合、今後矛盾なくプロセスが進められる可能性が確実に残っているかどうかを調べる。
        必要十分な解析をする必要がある。
        """
        # 全く True が無ければそもそもダメだね
        if not available_group.any():
            contradict = True
        else:
            contradict = False
            # available_group_for_plasmid = np.copy(self.available_group_for_plasmid)
            # print(available_group_for_plasmid.astype(int))
            # quit()
        if contradict:
            return None
        else:
            return available_group
    def remove_plasmid_connection(self, idx1, idx2):
        new_value1, new_value2 = self.remove_plasmid_connection_core(idx1, idx2)
        if (new_value1 != 0) and (new_value2 != 0):
            if (new_value1 == 1) or (new_value2 == 1):
                self.process_lonly_connection() # 新たに 1 が誕生しても、それが属するグループが決定していいなければ self.unfavorable_idx_pairs_kept には登録されないようになっている
            return True     # not processed
        else:
            if new_value1 == 0:
                available_group_1 = self.generate_available_group(idx1) # self.available_group_for_plasmid はまだアプデされない
                # グループを割り当てることが不可能な場合
                if available_group_1 is None:
                    self.undo_remove_plasmid_connection(idx1, idx2)
                    return False     # not processed
                else:
                    self.undo_remove_plasmid_connection(idx1, idx2)     # 直下の行の update_available_group_for_plasmid で統一的に書くため、一旦戻す
                    self.update_available_group_for_plasmid(idx1, available_group_1)
                    assert (self.available_group_for_plasmid.sum(axis=1) >= self.plasmid_loc_in_cluster_group.sum(axis=0)).all()
                    return True     # processed
            if new_value2 == 0:
                available_group_2 = self.generate_available_group(idx2) # self.available_group_for_plasmid はまだアプデされない
                # グループを割り当てることが不可能な場合
                if available_group_2 is None:
                    self.undo_remove_plasmid_connection(idx1, idx2)
                    return False     # not processed
                # グループを割り当てることが可能な場合、アプデを検討
                else:
                    self.undo_remove_plasmid_connection(idx1, idx2)     # 直下の行の update_available_group_for_plasmid で統一的に書くため、一旦戻す
                    self.update_available_group_for_plasmid(idx2, available_group_2)
                    assert (self.available_group_for_plasmid.sum(axis=1) >= self.plasmid_loc_in_cluster_group.sum(axis=0)).all()
                    return True     # processed
    def remove_plasmid_connection_core(self, idx1, idx2):
        # must be idx1 < idx2   (あえて assersion しないよ!)
        assert (idx1, idx2) not in self.unfavorable_idx_pairs_removed
        self.unfavorable_idx_pairs_removed.append((idx1, idx2))
        cluster_idx1 = self.cluster_idx_from_plasmid_idx[idx1]
        cluster_idx2 = self.cluster_idx_from_plasmid_idx[idx2]
        self.pattern_list_for_each_plasmid[cluster_idx1, idx2] -= 1
        self.pattern_list_for_each_plasmid[cluster_idx2, idx1] -= 1
        return self.pattern_list_for_each_plasmid[cluster_idx2, idx1], self.pattern_list_for_each_plasmid[cluster_idx1, idx2]
    def undo_remove_plasmid_connection(self, idx1, idx2):
        # must be idx1 < idx2   (あえて assersion しないよ!)
        self.unfavorable_idx_pairs_removed.remove((idx1, idx2))
        cluster_idx1 = self.cluster_idx_from_plasmid_idx[idx1]
        cluster_idx2 = self.cluster_idx_from_plasmid_idx[idx2]
        self.pattern_list_for_each_plasmid[cluster_idx1, idx2] += 1
        self.pattern_list_for_each_plasmid[cluster_idx2, idx1] += 1
    def update_available_group_for_plasmid(self, idx, available_group):
        self.update_available_group_for_plasmid_core(idx, available_group)
        # 事後処理（新たにできた lonly_connection を守る）
        print(self.unfavorable_idx_pairs_kept, "kept")
        print(self.unfavorable_idx_pairs_removed, "removed")
        self.process_lonly_connection()
    def update_available_group_for_plasmid_core(self, idx, available_group):
        plasmid_can_be_in_same_group_pre = self.available_group_for_plasmid[self.available_group_for_plasmid[:, idx], :].sum(axis=0)
        plasmid_can_be_in_same_group_post = self.available_group_for_plasmid[available_group, :].sum(axis=0)
        cluster_idx = self.cluster_idx_from_plasmid_idx[idx]
        # 前までコネクト可能だったが、このたびコネクト不可能になったペアを探索・削除
        for idx2, is_pair_lost in enumerate(plasmid_can_be_in_same_group_pre * np.logical_not(plasmid_can_be_in_same_group_post)):
            if is_pair_lost and (cluster_idx != self.cluster_idx_from_plasmid_idx[idx2]):    # 同じクラスタに入っているのは禁止
                self.remove_plasmid_connection_core(*self.sort_idx(idx, idx2))
        # アプデ
        self.available_group_for_plasmid[:, idx] = available_group
        # アプデによってさらなるアプデが可能かもしれない
        """
        # 単純な数独（完全に解かなくて良い）
        EXAMPLE (N_groups, N_clusters): 本当は dtype = bool
            1 1         1 0
            1 1     ->  1 0
            0 1         0 1
        EXAMPLE
            1 0 1       1 0 0
            1 1 0   ->  0 1 0
            0 0 1       0 0 1
        EXAMPLE
            0 1 1       0 1 0       0 1 0
            1 1 0   ->  1 1 0   ->  1 0 0
            1 1 0       1 1 0       1 0 0
            0 0 1       0 0 1       0 0 1
        """
        cluster = self.cluster_list[self.cluster_idx_from_plasmid_idx[idx]]
        available_group_sub_matrix = self.available_group_for_plasmid[:, cluster] # fancy indexing はビューではなくコピーが返される
        for row, row_sum in enumerate(available_group_sub_matrix.sum(axis=1)):
            # とある行の合計が 1 の場合、その行の中でどの列が 1 であるかを特定。
            if row_sum == 1:
                col = np.where(available_group_sub_matrix[row, :])[0]
                assert len(col) == 1
                col = col[0]    # 特定
                # 特定された列の合計が 2 以上の場合、アプデ可能である。
                if (available_group_sub_matrix[:, col]).sum() > 1:
                    plasmid_idx = cluster[col]
                    available_group_sub_matrix[:, col] = False
                    available_group_sub_matrix[row, col] = True
                    self.update_available_group_for_plasmid_core(plasmid_idx, available_group_sub_matrix[:, col])
                    break   # アプデ可能な他の列が残っていたとしても、直上の self.update_available_group_for_plasmid_core で再帰的に処理される
        # 列から処理する場合: 行の処理でいろいろアプデされてる可能性があるから、再度 available_group_sub_matrix を取得する。
        available_group_sub_matrix = self.available_group_for_plasmid[:, cluster]
        for col, col_sum in enumerate(available_group_sub_matrix.sum(axis=0)):
            # とある列の合計が 1 の場合、その列の中でどの行が 1 であるかを特定
            if col_sum == 1:
                row = np.where(available_group_sub_matrix[:, col])[0]
                assert len(row) == 1
                row = row[0]    # 特定
                # 特定された行の合計が 2 以上の場合、アプデ可能である。
                if (available_group_sub_matrix[row, :].sum() > 1):
                    # アプデする行を決定
                    for c, v in enumerate(available_group_sub_matrix[row, :]):
                        if v and (c != col):
                            break
                    else:
                        # if (available_group_sub_matrix[row, :].sum() > 1) より、ここに入るはずがない
                        raise Exception("error")
                    plasmid_idx = cluster[c]
                    available_group_sub_matrix[:, c] = False
                    available_group_sub_matrix[row, c] = True
                    self.update_available_group_for_plasmid_core(plasmid_idx, available_group_sub_matrix[:, c])
                    break   # 残っていても、直上の self.update_available_group_for_plasmid_core で再帰的に処理される
    def process_lonly_connection(self):
        unfavorable_idx_pairs_to_be_removed = []
        while True:
            unfavorable_idx_pairs_kept, unfavorable_idx_pairs_removed = self.process_lonly_connection_core()
            unfavorable_idx_pairs_to_be_removed.extend(unfavorable_idx_pairs_removed)
            if len(unfavorable_idx_pairs_kept) > 0:
                self.unfavorable_idx_pairs_kept.extend(unfavorable_idx_pairs_kept)
            else:
                break
        if len(unfavorable_idx_pairs_to_be_removed) > 0:
            for idx1, idx2 in list(set(unfavorable_idx_pairs_to_be_removed)):
                self.remove_plasmid_connection(idx1, idx2)
    def process_lonly_connection_core(self):
        unfavorable_idx_pairs_kept = []
        unfavorable_idx_pairs_removed = []
        # コネクションが 1 つだけのものを処理（所属すべきグループが決定している場合にのみ適応される: 決定していない場合、そもそもそのコネクションは使わないという選択肢もあるため）
        for plasmid_idx1, is_group_determined in enumerate(self.available_group_for_plasmid.sum(axis=0) == 1):
            # plasmid_idx1 が所属すべきグループが決定している場合
            if is_group_determined:
                for cluster_idx2, N_connection in enumerate(self.pattern_list_for_each_plasmid[:, plasmid_idx1] == 1):
                    # plasmid_idx1 とのコネクションが 1 つしかないプラスミドがあった場合
                    if N_connection == 1:
                        # そのコネクションは削除してはならない
                        executed_pair_count = 0
                        for plasmid_idx2 in self.cluster_list[cluster_idx2]:
                            plasmid_pair = self.sort_idx(plasmid_idx1, plasmid_idx2)
                            if plasmid_pair in self.unfavorable_idx_pairs_removed:
                                continue
                            else:
                                executed_pair_count += 1
                                if plasmid_pair not in self.unfavorable_idx_pairs_kept:
                                    unfavorable_idx_pairs_kept.append(plasmid_pair)
                        assert executed_pair_count == 1
                        # cluster_1 中の plasmid_idx1 以外のプラスミドについて、plasmid_idx2 との関係を断つ
                        for plasmid_idx_in_cluster1 in self.cluster_list[self.cluster_idx_from_plasmid_idx[plasmid_idx1]]:
                            if plasmid_idx_in_cluster1 == plasmid_idx1:
                                continue
                            else:
                                plasmid_pair = self.sort_idx(plasmid_idx_in_cluster1, plasmid_idx2)
                                """
                                self.process_lonly_connection で、どうしても _core が 2 度回ることが多々ある。
                                ここでself.remove_plasmid_connection_core(*plasmid_pair) してしまうと、その 2 回目の時に remove_plasmid_connection_core 内で 
                                AssertionError assert (idx1, idx2) not in self.unfavorable_idx_pairs_removed してしまう。
                                """
                                if plasmid_pair not in self.unfavorable_idx_pairs_removed:
                                    unfavorable_idx_pairs_removed.append(plasmid_pair)
        return list(set(unfavorable_idx_pairs_kept)), unfavorable_idx_pairs_removed
    # チェック関数
    def assert_available_group_for_plasmid(self):
        for idx in range(self.N_plasmids):
            available_group = self.generate_available_group(idx)
            print(idx, all(available_group == self.available_group_for_plasmid[:, idx]))
            # assert all(available_group == self.available_group_for_plasmid[:, idx])

def export_results(recommended_grouping: RecommendedGrouping, save_dir: Path):
    print("\nexporting results...")
    recommended_grouping.save(save_dir / RecommendedGrouping.file_name) # execute_pre_survey では必ず一度 save される (そのさい path が登録される) ので、再度 path を与える必要はない
    recommended_grouping.draw_heatmaps(save_dir, display_plot=False)
    print("export: DONE")





# -*- coding: utf-8 -*-

import io
import gc
import numpy as np
from tqdm import tqdm
from pathlib import Path
from typing import List
from collections import defaultdict

# my modules
from ..modules import msa
from ..modules import my_classes as mc
from ..modules import ref_query_alignment as rqa

__all__ = [
    "execute_alignment", 
    "normalize_scores_and_apply_threshold", 
    "execute_msa", 
]

#############
# ALIGNMENT #
#############
def execute_alignment(ref_seq_list:mc.MyRefSeq, my_fastq:mc.MyFastQ, param_dict, save_dir:Path=None):
    # load if there is intermediate data
    skip = False
    if save_dir is not None:
        intermediate_results_save_path = save_dir.parent / f"{my_fastq.combined_name_stem}.intermediate_results.ir"
        if intermediate_results_save_path.exists():
            intermediate_results = IntermediateResults()
            intermediate_results.load(intermediate_results_save_path)
            ref_seq_list_tmp = intermediate_results.assert_identity(ref_seq_list, my_fastq, param_dict)
            if ref_seq_list_tmp is not None:
                ref_seq_list = ref_seq_list_tmp
                result_dict = intermediate_results.result_dict
                print("alignment: SKIPPED\n")
                skip = True
    intermediate_results_save_path = save_dir / f"{my_fastq.combined_name_stem}.intermediate_results.ir"
    if not skip:
        ###########
        # Execute #
        ###########
        print("executing alignment...")
        result_dict = execute_alignment_core(ref_seq_list, my_fastq, param_dict)
        print("alignment: DONE\n")
        if save_dir is not None:
            if not intermediate_results_save_path.parent.exists():
                intermediate_results_save_path = Path.home() / intermediate_results_save_path.name
            intermediate_results = IntermediateResults(ref_seq_list, my_fastq, param_dict, result_dict)
            intermediate_results.path = intermediate_results_save_path
    intermediate_results.save(intermediate_results_save_path)
    return result_dict, ref_seq_list    # orders might have been changed during "assert_identity" step.

def execute_alignment_core(ref_seq_list, my_fastq, param_dict):
    my_optimized_aligner_list = [rqa.MyOptimizedAligner(ref_seq, param_dict) for ref_seq in ref_seq_list]
    fastq_len = len(my_fastq)
    result_dict = defaultdict(list)
    pbar = tqdm(my_fastq.items(), ncols=100, mininterval=0.1, leave=True, bar_format='{l_bar}{bar}{r_bar}', total=len(my_fastq))   # \033[32m
    for query_id, (query_seq, q_scores) in pbar:
        pbar.set_postfix_str(query_id[:50])
        # calc scores for each ref_seq
        query_seq = mc.MySeq(query_seq)
        query_seq_rc = query_seq.reverse_complement()
        for my_optimized_aligner in my_optimized_aligner_list:
            conserved_regions = my_optimized_aligner.calc_circular_conserved_region(query_seq, omit_too_long=True)
            conserved_regions_rc = my_optimized_aligner.calc_circular_conserved_region(query_seq_rc, omit_too_long=True)
            if conserved_regions is not None:
                my_result = my_optimized_aligner.execute_circular_alignment_using_conserved_regions(query_seq, conserved_regions)
            else:
                my_result = rqa.MyResult()
            if conserved_regions_rc is not None:
                my_result_rc = my_optimized_aligner.execute_circular_alignment_using_conserved_regions(query_seq_rc, conserved_regions_rc)
            else:
                my_result_rc = rqa.MyResult()

            # レジスター
            result_dict[query_id].extend([my_result, my_result_rc])
            gc.collect()
    return result_dict

class IntermediateResults(mc.MyTextFormat, mc.MyHeader):
    intermediate_results_version = "ir_0.2.0"
    def __init__(self, ref_seq_list=None, my_fastq=None, param_dict=None, result_dict=None) -> None:
        super().__init__()
        self.header += f"\nintermediate_results_version: {self.intermediate_results_version}"
        self.path = None
        self.keys = [
            ("header", "str"), 
            ("datetime", "str"), 
            ("ref_seq_names", "list"), 
            ("ref_seq_hash_list", "list"), 
            ("my_fastq_names", "list"), 
            ("my_fastq_hash", "str"), 
            ("param_dict", "dict"), 
            ("query_id_list", "list")
        ]
        self.non_default_keys_start_idx = 8
        self.param_dict_keys_matter = ['gap_open_penalty', 'gap_extend_penalty', 'match_score', 'mismatch_score']
        self.query_id_list = []
        if result_dict is not None:
            # my aligner related info
            self.ref_seq_names = [ref_seq.path.name for ref_seq in ref_seq_list]
            self.ref_seq_hash_list = [ref_seq.my_hash for ref_seq in ref_seq_list]
            self.my_fastq_names = [fastq_path.name for fastq_path in my_fastq.path]
            self.my_fastq_hash = my_fastq.my_hash
            self.param_dict = param_dict
            # result_dict related info
            assert len(result_dict) == len(my_fastq)
            assert result_dict.keys() == my_fastq.keys()
            idx = 0
            for query_id, result_list in result_dict.items():
                self.query_id_list.append(query_id)
                for my_result in result_list:
                    result_key = f"result{idx}"
                    setattr(self, result_key, my_result.to_dict())
                    self.keys.append((result_key, "dict"))
                    idx += 1
            assert len(my_fastq) * len(ref_seq_list) * 2 == idx # リバコン(rc) もあるので二倍で assertion
    def assert_identity(self, ref_seq_list, my_fastq, param_dict):
        ref_seq_names = [ref_seq.path.name for ref_seq in ref_seq_list]
        ref_seq_hash_list = [ref_seq.my_hash for ref_seq in ref_seq_list]
        my_fastq_names = [fastq_path.name for fastq_path in my_fastq.path]
        my_fastq_hash = my_fastq.my_hash
        is_param_dict_same = all(param_dict[k] == self.param_dict[k] for k in self.param_dict_keys_matter)
        if (self.ref_seq_names == ref_seq_names) and\
            (self.ref_seq_hash_list == ref_seq_hash_list) and\
            (self.my_fastq_names == my_fastq_names) and\
            (self.my_fastq_hash == my_fastq_hash) and is_param_dict_same:
            return ref_seq_list
        elif (set(self.ref_seq_names) == set(ref_seq_names)) and\
            (set(self.ref_seq_hash_list) == set(ref_seq_hash_list)) and\
            (self.my_fastq_names == my_fastq_names) and\
            (self.my_fastq_hash == my_fastq_hash) and is_param_dict_same:
            # intermediat_resultsに応じて順番を並べ直す
            return [ref_seq_list[ref_seq_hash_list.index(ref_seq_hash)] for ref_seq_hash in self.ref_seq_hash_list]
        else:
            return None
    def save(self, save_path):
        super().save(save_path, zip=True)
    def load(self, load_path):
        self.path = load_path
        self.keys = super().load(load_path, zip=True)
        for k, v in self.param_dict.items():
            try:
                self.param_dict[k] = int(v)
            except:
                self.param_dict[k] = float(v)
    @property
    def result_dict(self):
        result_dict = {}
        result_list = []
        ref_seq_idx = 0
        ref_seq_idx_max = len(self.ref_seq_names) * 2
        query_idx = 0
        for key, data_type in self.keys[self.non_default_keys_start_idx:]:
            assert data_type == "dict"
            my_result = rqa.MyResult()
            my_result.apply_dict_params(getattr(self, key))
            result_list.append(my_result)
            ref_seq_idx += 1
            if ref_seq_idx == ref_seq_idx_max:
                query_id = self.query_id_list[query_idx]
                result_dict[query_id] = result_list
                ref_seq_idx = 0
                query_idx += 1
                result_list = []
        assert ref_seq_idx == 0
        assert query_idx == len(self.query_id_list)
        return result_dict

##############
# ASSIGNMENT #
##############
def normalize_scores_and_apply_threshold(ref_seq_list, my_fastq, result_dict, param_dict):
    print("normalizing scores...")
    query_assignment = msa.QueryAssignment(ref_seq_list, my_fastq, result_dict)
    print("normalization: DONE\n")
    # set threshold for assignment
    query_assignment.set_assignment(param_dict["score_threshold"])
    return query_assignment

def draw_and_save_query_assignment(query_assignment:msa.QueryAssignment, save_dir, display_plot=False):
    print("drawing and saving figures...")
    query_assignment.save_scores(save_dir)
    query_assignment.draw_distributions(save_dir, display_plot=False)
    query_assignment.draw_alignment_score_scatters(save_dir, display_plot=False)
    query_assignment.draw_alignment_score_scatters_rotated(save_dir, display_plot=False)
    print("drawing and saving:DONE\n")

#################
# MSA/CONSENSUS #
#################
def execute_msa(result_dict, query_assignment:msa.QueryAssignment, param_dict):
    print("executing MSA...")
    my_msa_list = []
    msa.MyMSA.set_sbq_pdf(query_assignment.my_fastq)
    for ref_seq, my_fastq_subset, result_list in query_assignment.iter_assignment_info(result_dict):
        print(f"processing {ref_seq.path.name}...")
        my_msa_aligner = msa.MyMSAligner(ref_seq, my_fastq_subset, result_list)
        my_msa_list.append(my_msa_aligner.execute(param_dict))
    print("MSA: DONE\n")
    return my_msa_list

#############
# CONSENSUS #
#############
def export_results(my_msa_list: List[msa.MyMSA], save_dir):
    print("exporting results...")
    for my_msa in my_msa_list:
        my_msa.export_consensus_fastq(save_dir)
        my_msa.export_gif(save_dir)
        my_msa.export_consensus_alignment(save_dir)
    print("export: DONE")

def export_log(ref_seq_list:list, my_fastq:mc.MyFastQ, param_dict, query_assignment: msa.QueryAssignment, save_dir:Path):
    print("exporting log...")
    my_log = MyLog(ref_seq_list, my_fastq, param_dict, query_assignment)
    my_log.save(save_path = save_dir / f"{my_fastq.combined_name_stem}.log")
    print("export: DONE")

class MyLog(mc.MyTextFormat, mc.MyHeader):
    def __init__(self, ref_seq_list: list=None, my_fastq: mc.MyFastQ=None, param_dict: dict=None, query_assignment=None) -> None:
        super().__init__()
        if ref_seq_list is not None:
            self.input_reference_files = [refseq.path for refseq in ref_seq_list]
            self.input_fastq_files = [fastq_path for fastq_path in my_fastq.path]
            self.input_reference_hash_list = [refseq.my_hash for refseq in ref_seq_list]
            self.input_fastq_hash = my_fastq.my_hash
            self.alignment_params = param_dict
            self.score_matrix = self.get_score_matrix()
            self.assignment_summary = query_assignment.get_assignment_summary()
            self.error_matrix_with_prior, self.error_matrix_without_prior = self.get_error_matrices()
            self.class_attributes = self.get_class_attributes()
        # keys required for MyTextFormat
        self.keys = [
            ("header", "str"), 
            ("datetime", "str"), 
            ("input_reference_files", "listPath"), 
            ("input_fastq_files", "listPath"), 
            ("input_reference_hash_list", "list"), 
            ("input_fastq_hash", "str"), 
            ("alignment_params", "dict"), 
            ("score_matrix", "str"), 
            ("assignment_summary", "str"), 
            ("error_matrix_with_prior", "str"), 
            ("error_matrix_without_prior", "str"), 
            ("class_attributes", "dict"), 
        ]
    def get_score_matrix(self):
        my_custom_matrix = rqa.MyAlignerBase(self.alignment_params).my_custom_matrix
        return self.matrix2string(my_custom_matrix.matrix, digit=3, round=True)
    def get_error_matrices(self):
        P_N_dict_dict_with_prior, P_N_dict_dict_without_prior = msa.MyMSA.P_N_dict_dict_2_matrix(self.alignment_params)
        return (
            self.matrix2string(P_N_dict_dict_with_prior, bases=msa.MyMSA.bases, digit=None, round=False), 
            self.matrix2string(P_N_dict_dict_without_prior, bases=msa.MyMSA.bases, digit=None, round=False)
        )

    @staticmethod
    def get_class_attributes():
        class_attributes_dict = {}
        # alignment params
        class_attributes_dict.update(
            ALIGNMENT_default_repeat_max = rqa.MyOptimizedAligner.default_repeat_max, 
            ALIGNMENT_percentile_factor = rqa.MyOptimizedAligner.percentile_factor, 
            ALIGNMENT_omit_too_long = rqa.MyOptimizedAligner.omit_too_long, 
            ALIGNMENT_consecutive_true_threshold = rqa.WindowAnalysis.consecutive_true_threshold, 
        )
        # msa/consensus params
        class_attributes_dict.update(
            ALIGNMENT_algorithm_ver = rqa.MyAlignerBase.alignment_algorithm_version, 
            QUERY_assignment_version = msa.QueryAssignment.query_assignment_version, 
            MSALIGN_N_polish = msa.MyMSAligner.N_polish, 
            MSA_algorithm_version = msa.MyMSA.algorithm_version, 
            MSA_bases = msa.MyMSA.bases, 
            MSA_letter_code_dict = msa.MyMSA.letter_code_dict, 
            MSA_sbq_pdf_version = msa.MyMSA.sbq_pdf_version, 
        )
        # consensus file_format info
        class_attributes_dict.update(
            FORMAT_file_format_version = msa.MyMSA.file_format_version, 
            FORMAT_ref_seq_related_save_order = msa.MyMSA.ref_seq_related_save_order,     # [tuple, tuple, ...]
            FORMAT_query_seq_related_save_order = msa.MyMSA.query_seq_related_save_order,   # [tuple, tuple, ...]
            FORMAT_size_indicator_bytes = msa.MyByteStr.size_indicator_bytes, 
            FORMAT_size_format = msa.MyByteStr.size_format, 
            FORMAT_my_bit = msa.MyByteStr.MyBIT.my_bit, 
            FORMAT_struct_bit = msa.MyByteStr.MyBIT.struct_bit, 
            FORMAT_seq_offset = msa.MyByteStr.MyBIT.seq_offset, 
        )
        return class_attributes_dict
    @staticmethod
    def matrix2string(matrix, bases="ATCG", digit=3, round=True):
        bio = io.BytesIO()
        if round:
            np.savetxt(bio, matrix, fmt=f"%{digit}d")
        else:
            np.savetxt(bio, matrix, fmt=f"%.5e")
            digit = 11
        matrix_str = bio.getvalue().decode('latin1')
        bases = bases + "*"
        output = (
            " " * (digit + 1)
            + (" " * digit).join(b for b in bases)
        )
        for b, m in zip(bases, matrix_str.split("\n")):
            output += f"\n{b} {m}"
        return output






# -*- coding: utf-8 -*-

import copy
import argparse

from . import *
from ..modules import my_classes as mc

if __name__ == "__main__":
    """
    python -m savemoney.pre_survey path_to_plasmid_map_dir save_dir_base
    """

    param_dict = copy.deepcopy(default_pre_survey_param_dict)

    # パーサーの設定
    parser = argparse.ArgumentParser()
    parser.add_argument("plasmid_map_dir_paths", help="path to plasmid map_directory", type=str)
    parser.add_argument("save_dir_base", help="save directory path", type=str)
    for key, val in param_dict.items():
        parser.add_argument(f"-{mc.key2argkey(key)}", help=f"{key}, optional, default_value = {val}", type=type(val), default=val)

    # 取得した引数を適用
    args = parser.parse_args()
    args_dict =vars(args)
    for key in param_dict.keys():
        param_dict[key] = args_dict[mc.key2argkey(key)]

    pre_survey(args.plasmid_map_dir_paths, args.save_dir_base, **param_dict)




# -*- coding: utf-8 -*-

import copy
import argparse

from . import *
from ..modules import msa
from ..post_analysis.post_analysis import default_post_analysis_param_dict

if __name__ == "__main__":
    """
    python -m savemoney.show_consensus
    """

    param_dict = copy.deepcopy(msa.MyMSA.default_print_options)

    # パーサーの設定
    parser = argparse.ArgumentParser()
    parser.add_argument("consensus_alignment_path", help="path to consensus_alignment (*.ca) file", type=str)
    for key, val in param_dict.items():
        parser.add_argument(f"--{key}", help=f"{key}, optional, default_value = {val}", type=type(val), default=val)

    # 取得した引数を適用
    args = parser.parse_args()
    args_dict =vars(args)
    for key in param_dict.keys():
        param_dict[key] = args_dict[key]

    show_consensus(args.consensus_alignment_path, **param_dict)




# -*- coding: utf-8 -*-

from pathlib import Path

from ..modules import msa
from ..post_analysis.post_analysis import default_post_analysis_param_dict

def ca2bam(consensus_alignment_path: str):
    my_msa = msa.MyMSA(param_dict=default_post_analysis_param_dict)
    my_msa.load_consensus_alignment(consensus_alignment_path)
    my_msa.convert_to_bam(save_dir=Path(consensus_alignment_path).parent, ext_not_exported=[".sorted.sam", ".sorted.bam", ".sorted.bam.bai"])

